#include "BlockMerge.h"
#include <unordered_set>
#include <vector>

#define DEBUG_SWITCH_bm 0
#if DEBUG_SWITCH_bm
#define print(str) std::cout << str << "\n";
#else
#define print(str) //
#endif

using namespace std;

void BlockMerge::execute() {
    print("============ start BlockMerge ============");
    change = 0;
    for (auto func : unit->get_functions()) {
        findBLocks(func);
    }
    print("============ end BlockMerge ============");
}

bool phiCheck(BasicBlock *bb,BasicBlock* start) {
    auto func = bb->getParent();
    print("*** entry = "<<func->getEntry()->getNo()<<" "<<start->getNo()<<" ***");
    auto bblist = func->getBlockList();
    for (auto instr = bb->begin(); instr != bb->end();
         instr = instr->getNext()) {
        if (!instr->isPhi())
            break;
        auto phi = dynamic_cast<PhiInstruction *>(instr);
        print("phi "<<phi->getDef()->toStr());
        int cnt = 0;

        auto phiBBs = phi->getPhiBBs();
        auto phiOperands = phi->getPhiOperands();

        std::set<BasicBlock*> removebbs;
        int old = -1;
         for (auto src:phi->getSrcs()) {
            print("PHIBB "<<src.first->getNo()<<" "<<src.second->toStr());
            if (std::find(bblist.begin(), bblist.end(), src.first) !=
                bblist.end()) {
                cnt++;
            } else {
                removebbs.insert(src.first);
            }
        }
        print("cnt = "<<cnt);
        for(auto b:removebbs){
            print("remove bb = "<<b->getNo()<<" "<<phi->getSrc(b)->toStr());
            phi->getSrcs().erase(b);
        }
        removebbs.clear();
        // BasicBlock*srcbb=nullptr;
        // for (int j = 0; j < phi->getSrcs.size(); j++) {
        //     print("PHIBB "<<phiBBs[j]->getNo());
        //     if (std::find(bblist.begin(), bblist.end(), phiBBs[j]) !=
        //         bblist.end()) {
        //         cnt++;
        //         srcbb=phiBBs[j];
        //     } else {
        //         old = j;
        //     }
        // }
        // if (old != -1) {
        //     phi->getSrcs().erase(phiBBs[old]);
        //     phi->getPhiOperands_v().erase(phi->getPhiOperands_v().begin() +
        //                                   old);
        //     phi->getPhiBBs_v().erase(phi->getPhiBBs_v().begin() + old);
        // }
        // print(srcbb->getNo());
        if(func->getEntry()==start)return 0;

        // for (auto src : phi->getSrcs()) {
        //     if (std::find(bblist.begin(), bblist.end(), src.first) !=
        //         bblist.end()) {
        //         cnt++;
        //     }
        // }
        if (cnt >= 2)
            return 0;
    }
    return 1;
}

bool no_phi_src_entry(BasicBlock *bb, BasicBlock *start) {
    auto head = start->end();
    for (auto instr = head->getNext(); instr != head;
         instr = instr->getNext()) {
        print(instr->getInstName());
        if (instr->isPhi()) {
            auto phi = dynamic_cast<PhiInstruction *>(instr);
            for (auto src : phi->getSrcs()) {
                if (src.first == start)
                    return false;
            }
        }
    }
    return true;
}

void BlockMerge::findBLocks(Function *func) {
    print("===== findBLocks =====");
    std::set<BasicBlock *> visitedBlock = {};
    for (auto bb : func->getBlockList()) {
        if (std::find(visitedBlock.begin(), visitedBlock.end(), bb) !=
            visitedBlock.end())
            continue;
        ;
        print("*********************check    bb: " << bb->getNo());

        if (findinst(Instruction::COND, bb) != nullptr
            // || findinst(Instruction::CMP, bb) != nullptr
            // || findinst(Instruction::PHI, bb) != nullptr
        )
            continue;
        if (bb->getNumOfSucc() != 1
            // || bb->getInstSum() <= 1
        )
            continue;

        print("BB: " << bb->getNo());
        mergeList.clear();
        BasicBlock *block = bb;
        while (true) {

            if (findinst(Instruction::COND, block))
                break;
            if (block->getNumOfSucc() < 1)
                break;

            visitedBlock.insert(block);
            auto succ = block->getSuccBlock()[0];
            if (!phiCheck(succ,bb))
                break;
            print("  succ: " << succ->getNo()
                             << " predsNum = " << succ->getNumOfPred());
            if (succ->getNumOfPred() == 1 
                // !(func->getEntry() == bb && no_phi_src_entry(succ,block))
                &&!findinst(Instruction::ALLOCA, succ)
                // &&!findinst(Instruction::PHI, succ)
                //  && succ->getInstSum() > 1
            ) {
                // mergeList.insert(succ);
                visitedBlock.insert(succ);
                mergeList.push_back(succ);
                block = succ;
            } else {
                break;
            }
        }
        if (mergeList.size() > 0) {
            merge(func, bb);
            change = 1;
        }
    }
    visitedBlock.clear();
}

void BlockMerge::printInsts(BasicBlock *start) {
    auto head = start->end();
    for (auto instr = head->getNext(); instr != head;
         instr = instr->getNext()) {
        print(instr->getInstName());
    }
    print("== new bb insts ==\n");
}

void BlockMerge::merge(Function *func, BasicBlock *start) {
    print("== start merge ==");

    auto uncond = findinst(Instruction::UNCOND, start);
    start->remove(uncond);
    // for (auto it = mergeList.rbegin(); it != mergeList.rend(); ++it) {
    //     auto bb=(*it);
    // BasicBlock *last = start;
    for (auto bb : mergeList) {
        std::vector<Instruction *> mergeInst = {};
        print("  merge bb: " << bb->getNo());
        // for (auto instr = bb->begin(); instr != bb->end();
        //      instr = instr->getNext()) {
        auto head = bb->end();
        for (auto instr = head->getNext(); instr != head;
             instr = instr->getNext()) {
            if (bb != mergeList.back() && instr->isUncond())
                continue;
            print(instr->getInstName());
            // mergeInst.insert(instr);
            mergeInst.push_back(instr);

            // print("startBB = " << start->getNo() << " bb = " << bb->getNo());
            // // bb->remove(instr);
            // // start->insertBack(instr);
            // start->end()->getPrev()->setNext(instr);
            // instr->setPrev(start->end()->getPrev());
            // instr->setNext(start->end());
            // start->end()->setPrev(instr);

            // instr->setParent(start);
        }
        replacePhiBB(bb, start);
        for (auto instr : mergeInst) {
            print("------------------ " << instr->getInstName());
            bb->remove(instr);
            start->insertBack(instr);
        }

        print("\n== new bb " << start->getNo() << " insts ==");

        // printInsts(start);

        print("start maintain pred&succ");
        // if (bb->existsInPred(last))
        //     bb->removePred(last);
        std::set<BasicBlock *> succ_tmp;
        for (auto succ : bb->getSuccBlock()) {
            succ_tmp.insert(succ);
        }
        // if (start->existsInSucc(bb))
        //     start->removeSucc(bb);
        bb->deleteAllRelation();
        func->remove(bb);
        for (auto succ : succ_tmp) {
            if (succ == start)
                continue;
            print("start == succ: " << start->getNo() << succ->getNo());
            replacePhiBB(succ, start);
            if (!start->existsInSucc(succ))
                start->addSucc(succ);
            if (!succ->existsInPred(start))
                succ->addPred(start);
        }
        // last = bb;
    }
    print("== merge end ==");
    auto cond = findinst(Instruction::COND, start);
    if (cond &&
        dynamic_cast<CondBrInstruction *>(cond)->getTrueBranch() == start) {
        print("find self spin!");
        // 处理phi指令
        replacePhiBB(start, start);
        if (!start->existsInSucc(start))
            start->addSucc(start);
        if (!start->existsInPred(start))
            start->addPred(start);
    }
}

void BlockMerge::replacePhiBB(BasicBlock *succ, BasicBlock *start) {
    // for (auto pred : succ->getPredBlock()) {
    //     print("succ: " << succ->getNo() << " pred " << pred->getNo());
    // }
    print("replaecBB");
    auto head = succ->end();
    for (auto instr = head->getNext(); instr != head;
         instr = instr->getNext()) {
        if (!instr->isPhi())
            continue;
        PhiInstruction *phi = dynamic_cast<PhiInstruction *>(instr);
        // auto phiBBs = phi->getPhiBBs();
        // auto phiOperands = phi->getPhiOperands();

        // Fix: 处理多src phi时可能产生问题
        // BasicBlock *old = nullptr;

        std::set<BasicBlock*> changebbs;
        for(auto src:phi->getSrcs()){
            if (!succ->existsInPred(src.first)&&src.first!=start) {
                changebbs.insert(src.first);
            }
        }
        for(auto b:changebbs){
            phi->getSrcs()[start] = phi->getSrc(b);
            phi->getSrcs().erase(b);
        }
        // int old = -1;
        // for (int j = 0; j < phiBBs.size(); j++) {
        //     if (!succ->existsInPred(phiBBs[j])) {
        //         old = j;
        //     }
        // }
        // print("old = " << old);
        // if (old != -1) {
        //     phi->getSrcs().erase(phiBBs[old]);
        //     phi->getSrcs()[start] = phiOperands[old];
        //     phi->getPhiOperands_v().erase(phi->getPhiOperands_v().begin() +
        //                                   old);
        //     phi->getPhiOperands_v().push_back(phiOperands[old]);
        //     phi->getPhiBBs_v().erase(phi->getPhiBBs_v().begin() + old);
        //     phi->getPhiBBs_v().push_back(start);
        // }

        // for (auto src : phi->getSrcs()) {
        //     if (!start->existsInPred(src.first)) {
        //         // src.first=start;
        //         phi->getSrcs()[start] = src.second;
        //         phi->getSrcs().erase(src.first);
        //     }
        // }
    }
}