#include "ConstantUnfold.h"
#include "Instruction.h"
#include "Unit.h"
#include <unordered_set>
#define ITERBOUND 40
#define DEBUG0805

#define DEBUG_SWITCH_cu 0
#if DEBUG_SWITCH_cu
#define print(str) std::cout << str << "\n";
#else
#define print(str) //
#endif

// #define DEBUG0429

void ConstantUnfold::pass() {
    detector = new LoopDetection(unit);
    detector->run();
    for (auto it = detector->begin(); it != detector->end(); it++) {

        auto oneloop = *it;
        checkLoop(oneloop, (*(oneloop->begin()))->getParent());
    }
}

// void ConstantUnfold::pass() {
// #ifdef DEBUG0429
//     print("ConstantUnfold::pass()");
// #endif
//     std::vector<Function *> func_list = unit->get_functions();
// #ifdef DEBUG0429
//     print("func_list.size() = " << func_list.size() << "");
// #endif
//     for (auto func = func_list.begin(); func != func_list.end(); func++) {
//         checkLoop(*func);
//     }
//     // for(int )
//     // auto loops = func->loop_begin();
// }

void ConstantUnfold::checkLoop(blockSet *oneloop, Function *func) {
    print("======checkLoop======= " << func->getSymPtr()->toStr()
                                    << " =============");

    // 一个关于cond块的后继块是不是body块的测试
    // 检查body块中是否存在call指令
    BasicBlock *body_bb =
        *(oneloop->begin()); // 这里zkd代码是用一个循环找的，为什么？
    if (body_bb->getSuccBlock().size() != 2) {

        return;
    }
    bool hasCall = false;
    for (Instruction *ins = body_bb->begin(); ins != body_bb->end();
         ins = ins->getNext()) {
        if (dynamic_cast<CallInstruction *>(ins))
            hasCall = false;
    }
    if (hasCall) {
        return;
    }
    int phiCount = 0;
    BasicBlock *while_bb = *(oneloop->begin());
    for (Instruction *ins = while_bb->begin(); ins != while_bb->end();
         ins = ins->getNext()) {
        if (ins->isPhi() == true)
            phiCount++;
    }
    if (phiCount != 1) {
        return;
    }
    unfold(oneloop, func);
}

void ConstantUnfold::pass2() {
    // printf("condstantunfold pass2");
    // 接下来处理内层循环变量的终值
    // 在cond_bb中找到icmp指令
    // printf("condbb label = %d\n", condbb->getNo());
    // printf("bodybb label = %d\n", bodybb->getNo());
    // condbb.
    BasicBlock *mtempbb = condbb->getPredBlock()[0];
    // printf("mtempbb label = %d\n", mtempbb->getNo());
    for (Instruction *ins = bodybb->begin(); ins != bodybb->end();
         ins = ins->getNext()) {
        // printf("ins = %d\n", ins);
        // printf("ins = %s\n", ins->getInstName().c_str());
        if (ins->isLoad()) {
            // printf("ins op0 = %d\n", ins->getOp()[0]->getLabel());
            // printf("ins op1 = %d\n", ins->getOp()[1]->getLabel());
            // printf("ins op2 = %d\n", ins->getOp()[2]->getLabel());
            // printf("ins op3 = %d\n", ins->getOp()[3]->getLabel());
        }
    }
}
void ConstantUnfold::unfold(loop_item *oneloop, Function *func) {}
// printf("is sitofp = %d\n", ins->isSitofp());
void ConstantUnfold::unfold(blockSet *oneloop, Function *func) {
    print("============= " << func->getSymPtr()->toStr() << " =============");
    print("in unfold func first line");
    // BasicBlock* cond_bb = oneloop->cond_bb;
    // BasicBlock* body_bb = oneloop->while_bb;
    // BasicBlock* end_bb = oneloop->end_bb;

    // z0804 新的循环查找
    BasicBlock *body_bb = *(oneloop->begin());
    BasicBlock *cond_bb = detector->getCondbb(body_bb);
    BasicBlock *end_bb = detector->getEndbb(body_bb);

    this->condbb = cond_bb;
    this->bodybb = body_bb;
    this->endbb = end_bb;

    print("|| cond_bb = " << cond_bb->getNo() << "");
    print("|| body_bb = " << body_bb->getNo() << "");
    print("|| end_bb = " << end_bb->getNo() << "");
    // cond_bb->removeSucc(body_bb);
    // body_bb->removePred(cond_bb);

    int start, end;
    Instruction *br = (body_bb->getBrorRet());
    Instruction *phi = body_bb->begin();
    if (!phi->isPhi())
        return;
    print("before assert cond bb dose not start with phi");
    assert(phi->isPhi() && "cond bb dose not start with phi");
    auto operands = (dynamic_cast<PhiInstruction *>(phi))->getPhiOperands();
    auto bbs = (dynamic_cast<PhiInstruction *>(phi))->getPhiBBs();
    if (operands.size() != 2) {
        return;
    }
    if (!operands[0]->isConst()) {
        return;
    }
    assert(bbs[0] != body_bb && "Inheriant const from loop by phi");

    start = operands[0]->getConstVal();
    print("start1 = " << start);
    Instruction *iter = operands[1]->getDef();
    Operand *myop = dynamic_cast<CondBrInstruction *>(br)->getCond();
    print("myop = " << myop->toStr());
    assert(myop != nullptr && "myop == nullptr");
    Instruction *comparer = myop->getDef();
    // std::vector<Operand*> ops = br->getOp();
    assert(comparer != nullptr && "br->getOp() == nullptr");
    if (!comparer->isCmp()) {
        return;
    }
    print(comparer->getOp()[2]->toStr());
    // return;
    if (comparer->getOp().size() >= 3 && comparer->getOp()[2] &&
        !comparer->getOp()[2]->isConst()) {
        return;
    }
    end = comparer->getOp()[2]->getConstVal();
    // // may cause problem!!
    Operand *op1 = comparer->getOp()[1];
    Operand *op2 = op1->getDef()->getOp()[1];
    print("op1 = " << op1->toStr() << " "
                   << op1->getDef()->getOp()[1]->toStr());
    print("phi->getDef() = " << phi->getDef()->toStr());
    if (op2 != phi->getDef()) {
#ifdef DEBUG0429_terminal
        print("its no = " << comparer->getOp()[1]->getLabel() << "");
        print("its no = " << op2->getLabel() << "");
        print("comparer->getOp()[0] != phi");
#endif
        return;
    }
#ifdef DEBUG0429
#endif
    print("stage2");
    // return;
    int iterPattern = -1;
    int accumulator;
    if (iter->getOp().size() != 3 || !iter->getOp()[2]->isConst()) {
        return;
    }
    accumulator = iter->getOp()[2]->getConstVal();
    print("accumulator = " << accumulator << "");

    if (iter->isBin() && iter->getOpcode() == BinaryInstruction::ADD) {
        iterPattern = 0;
    } else if (iter->isBin() && iter->getOpcode() == BinaryInstruction::SUB) {
        iterPattern = 1;
        accumulator = -accumulator;
    } else {
        return;
    }
    int iternum = (end - start) / accumulator;
    iternum = abs(iternum);
    print("iternum  = " << iternum);
    if (iternum > ITERBOUND) {
#ifdef DEBUG0429_terminal
        print("(end - start) / accumulator > ITERBOUND");
#endif
        return;
    }
    if (iterPattern == -1) {
#ifdef DEBUG0429_terminal
        print("iterPattern == -1");
#endif
        return;
    }

    if (comparer->getOpcode() != CmpInstruction::L) {
#ifdef DEBUG0429_terminal
        print("comparer->getOpcode() != CmpInstruction::L");
#endif
        return;
    }
#ifdef DEBUG0429
#endif
    // std::cout<<"stage3\n";
    std::unordered_set<Instruction *> ignored_ins{phi, comparer, br};
#ifdef DEBUG0429
    print("ignored: phi = " << phi << " comparer = " << compare) r
        << " br = " << br << "\n";
#endif
    BasicBlock *outBB = detector->getEndbb(body_bb);
// for(auto BB : cond_bb->getSuccBlock())
// {
//     if(BB != body_bb)outBB = BB;
// }
#ifdef DEBUG0429
    print("outBB.no = " << outBB->getNo() << "");
#endif

    auto block_list0 = func->getBlockList();

    BasicBlock *new_bb = new BasicBlock(func);
    // std::cout<<"new_bb No = "<<new_bb->getNo()<<"\n";
    std::unordered_map<Instruction *, Instruction *> old2new;
    print("start = " << start << " end = " << end << "");

    // std::cout<<"unfold begin\n";
    std::vector<Function *> func_list = unit->get_functions();

    std::vector<Instruction *> sitofps;
    for (int h = 0; h < iternum; h++) {
        old2new.clear();

        print("below transfer body_bb for insert");

        // std::cout<<"iter1 = "<<iter<<"\n";
        //  body_bb->insertBefore(iter, comparer);
        //  body_bb->remove(iter);
        //  iter = comparer->getPrev();

        // std::cout<<"iter2 = "<<iter<<"\n";
        // 把外层循环i的值存回栈里
        // new StoreInstruction(dst, src, new_bb);
        // printf("body_bb next gep:");
        // for(int i = 0; i < body_bb->begin()->getNext()->getOp().size(); i ++)
        // {
        //     printf("operand%d = %d\n", i,
        //     body_bb->begin()->getNext()->getOp()[i]->getLabel());
        // }
        // 找到body块中的fsub指令，确认一下是否在copy时没有拷贝上浮点数信息
        Instruction *fsub = nullptr;
        Instruction *mysitofp = nullptr;
        for (Instruction *ins = body_bb->begin(); ins != body_bb->end();
             ins = ins->getNext()) {
            // if(ins->isGep())
            // //else if(ins->isCall())
            // {
            //     // std::cout<<"isgep iscall 1\n";
            //     Operand* d = ins->getOp()[0];
            //     printf("0710: gep1 operands");
            // if(ins->isSitofp() && ins->getOp()[0]->getLabel() == 68)
            // {
            //     mysitofp = ins;
            //     printf("mysitofp = ins;");
            // }
            // printf("ins = %s\n", ins->getInstName().c_str());
            // printf("ins opcode = %d\n", ins->getOpcode());
            // for(int i = 0; i < ins->getOp().size(); i ++)
            // {
            //     printf("operand%d = %d\n", i, ins->getOp()[i]->getLabel());
            //     printf("operand%d type = %s\n", i,
            //     ins->getOp()[i]->getType()->toStr().c_str());
            // }
            if (ins->isBin() && ins->getOpcode() == 0) {
                fsub = ins;
            }
            // }
            // printf("is sitofp(out) = %d\n", ins->isSitofp());
            // if(ins == iter)std::cout<<"is iter1\n";
            // if(ins->isGep())std::cout<<"is gep1\n";
            // if(ins->isCall())
            // {
            //     std::cout<<"call dst = "<<ins->getDef()->getLabel()<<" call
            //     dst =
            //     "<<dynamic_cast<CallInstruction*>(ins)->dst->getLabel()<<"\n";
            // }
            if (ignored_ins.find(ins) != ignored_ins.end()) {
                continue;
            }
            if (ins == iter || ins->isCond() || ins->isUncond()) {
                continue;
            }
            // std::cout<<"??? instr name = "<<ins->getInstName()<<"\n";
            Instruction *new_ins = ins->copy(ins, new_bb);
            // std::cout<<"new>>>\n";
            old2new.insert({ins, new_ins});
            // std::cout<<"insert>>>\n";
            if (new_ins->isSitofp()) {
                // printf("insert to sitofps");
                sitofps.push_back(new_ins);
            }

            // printf("new ins from copy");
            // printf("new ins = %s\n", old2new[ins]->getInstName().c_str());
            // printf("new ins opcode = %d\n", old2new[ins]->getOpcode());
            // for(int i = 0; i < old2new[ins]->getOp().size(); i ++)
            // {
            //     printf("operand%d = %d\n", i,
            //     old2new[ins]->getOp()[i]->getLabel()); printf("operand%d type
            //     = %s\n", i,
            //     old2new[ins]->getOp()[i]->getType()->toStr().c_str());
            // }
        }

        // printf("mysitofp = %d\n", mysitofp);
        // 看看fsub转换成=的指令是不是浮点指令
        // printf("fsub = %d\n", fsub);
        // printf("fsub insname 1 = %s\n", fsub->getInstName().c_str());
        // printf("fsub dst op label = %d\n", fsub->getOp()[0]->getLabel());
        // printf("fsub op0 type 1 = %s\n",
        // old2new[fsub]->getOp()[0]->getType()->toStr().c_str()); printf("old
        // fsub op0 type 1 = %s\n",
        // fsub->getOp()[0]->getType()->toStr().c_str());

        // print("below transfer body_bb for exchange operand");
        bool flag = false;

        for (Instruction *ins = body_bb->begin(); ins != body_bb->end();
             ins = ins->getNext()) {
            // print("=    "<<ins->getInstName());
            if (ignored_ins.find(ins) != ignored_ins.end()) {
                continue;
            }
            if (ins == iter) {
                start += accumulator;
                // std::cout<<"start2 = "<<start<<"\n";
                continue;
            } else if (ins->isCond() || ins->isUncond()) {
                continue;
            } else if (ins->isPhi()) {
                continue;
            } else if (ins->isGep())
            // else if(ins->isCall())
            {
                // std::cout<<"isgep iscall 1\n";
                Operand *d = ins->getOp()[0];
                Type *type = TypeSystem::intType;
                auto tmp = dynamic_cast<PointerType *>(d->getType());
                d = new Operand(new TemporarySymbolEntry(
                    new PointerType(tmp->getType()), SymbolTable::getLabel()));
                old2new[ins]->setOperand(0, d);
                // std::cout<<"isgep iscall 2\n";
            }

            // print("ins op size = " << ins->getOp().size() << "");
            // print(ins->getPrev()->getDef()->toStr());
            // print(ins->getOp()[0]->toStr());
            // print(ins->getPrev()->getDef()->getUse().size());
            // print(ins->getPrev()->getDef()->getUse()[0]->getInstName());
            if (ins->isStore() && ins->getPrev()->isGep() &&
                (ins->getPrev()->getDef()) == ins->getOp()[0]) {
                old2new[ins]->setOperand(0, old2new[ins->getPrev()]->getDef());
            } else if (ins->isLoad() && ins->getPrev()->isGep() &&
                       (ins->getPrev()->getDef()->getUse()[0]) == ins) {
                // old2new[ins]->setOperand(0,
                // old2new[ins->getPrev()]->getDef());
                Type *tmptype = ins->getOp()[1]->getType();
                if (!tmptype->isPtr()) {
                    assert(0 && "load operand1(src) type is not PTR!");
                }
                // dynamic_cast<PointerType*>(tmptype)->getType()->isFloat();
                if ((dynamic_cast<PointerType *>(tmptype))
                        ->getType()
                        ->isInt()) {
                    tmptype = TypeSystem::intType;
                } else if ((dynamic_cast<PointerType *>(tmptype))
                               ->getType()
                               ->isFloat()) {
                    // printf("here!");
                    tmptype = TypeSystem::floatType;
                }
                Operand *newope = new Operand(
                    new TemporarySymbolEntry(tmptype, SymbolTable::getLabel()));
                Operand *oldope = old2new[ins]->getOp()[0];
                // 拷贝def/use关系
                newope->setDef(ins);
                newope->addUse(old2new[ins->getNext()]);

                old2new[ins]->setOperand(0, newope);
            }

            for (unsigned i = 1; i < ins->getOp().size(); i++) {
                if (ins->isGep() && i == 1) {
                    if (ins->getOp()[1]->getDef() == nullptr)
                        continue;
                    if (ins->getOp()[1]->getDef()->isGep()) {
                        // std::cout<<"gep but not the first\n";
                        old2new[ins]->setOperand(
                            1, old2new[ins->getOp()[1]->getDef()]->getOp()[0]);
                    }
                }

                Instruction *ope = ins->getOp()[i]->getDef();
                if (ope == nullptr) {
                    // std::cout<<"continue\n";
                    continue;
                }
                if (ins->getOp()[i] == phi->getDef() || ope == iter) {
                    // if(ins->isSitofp())
                    //     std::cout<<"branch 1\n";
                    // old2new[ins]->setOperand(i, new Operand(new
                    // ConstantSymbolEntry(TypeSystem::floatType, start)));
                    old2new[ins]->setOperand(
                        i, new Operand(new ConstantSymbolEntry(
                               TypeSystem::constIntType, start)));
                } else if (old2new.find(ope) != old2new.end()) {
                    // print("branch old2new.find(ope) != old2new.end()");
                    // if(ins->isSitofp())
                    //     std::cout<<"branch2\n";
                    // printf("is sitofp = %d\n", ins->isSitofp());
                    // std::cout<<"setOperand 4\n";
                    old2new[ins]->setOperand(i, old2new[ope]->getOp()[0]);
                    if (ins->isStore() && i == 1 && ins->getPrev()->isGep()) {
                        // printf("gep before store op1 type = %s\n",
                        // ins->getPrev()->getOp()[1]->getSymbolEntry()->getType()->toStr().c_str());
                        assert(ins->getPrev()
                                   ->getOp()[1]
                                   ->getSymbolEntry()
                                   ->getType()
                                   ->isPtr() &&
                               "gep op1 is not pointer type!! will cause "
                               "nullptr fault!!");
                        // printf("isfloat? %d\n",
                        // dynamic_cast<PointerType*>(ins->getPrev()->getOp()[1]->getSymbolEntry()->getType())->getType()->isFloat());
                        PointerType *mytype =
                            dynamic_cast<PointerType *>(ins->getPrev()
                                                            ->getOp()[1]
                                                            ->getSymbolEntry()
                                                            ->getType());
                        if (mytype->isPtr()) {
                            Type *mytype2 = mytype->getType();
                            if (mytype2->isArray()) {
                                // printf("in array branch");
                                if (dynamic_cast<ArrayType *>(mytype2)
                                        ->getElementType()
                                        ->isFloat()) {
                                    // printf("in array1 branch");
                                    old2new[ins]
                                        ->getOp()[1]
                                        ->getSymbolEntry()
                                        ->setType(TypeSystem::floatType);
                                } else if (dynamic_cast<ArrayType *>(mytype2)
                                               ->getElementType()
                                               ->isInt()) {
                                    // printf("in array2 branch");
                                    old2new[ins]
                                        ->getOp()[1]
                                        ->getSymbolEntry()
                                        ->setType(TypeSystem::intType);
                                    // old2new[ins]->getOp()[1]->getSymbolEntry()->setType(TypeSystem::floatType);
                                } else {
                                    printf("!!!gep before store op1 type's "
                                           "type = %s\n",
                                           dynamic_cast<PointerType *>(
                                               ins->getPrev()
                                                   ->getOp()[1]
                                                   ->getSymbolEntry()
                                                   ->getType())
                                               ->getType()
                                               ->toStr()
                                               .c_str());
                                    assert(0 &&
                                           "not int type and not float type");
                                }
                            } else {

                                if (dynamic_cast<PointerType *>(
                                        ins->getPrev()
                                            ->getOp()[1]
                                            ->getSymbolEntry()
                                            ->getType())
                                        ->getType()
                                        ->isFloat()) {
                                    old2new[ins]
                                        ->getOp()[1]
                                        ->getSymbolEntry()
                                        ->setType(TypeSystem::floatType);
                                } else if (dynamic_cast<PointerType *>(
                                               ins->getPrev()
                                                   ->getOp()[1]
                                                   ->getSymbolEntry()
                                                   ->getType())
                                               ->getType()
                                               ->isInt()) {
                                    old2new[ins]
                                        ->getOp()[1]
                                        ->getSymbolEntry()
                                        ->setType(TypeSystem::intType);
                                    // old2new[ins]->getOp()[1]->getSymbolEntry()->setType(TypeSystem::floatType);

                                } else {
                                    printf("!!!gep before store op1 type's "
                                           "type = %s\n",
                                           dynamic_cast<PointerType *>(
                                               ins->getPrev()
                                                   ->getOp()[1]
                                                   ->getSymbolEntry()
                                                   ->getType())
                                               ->getType()
                                               ->toStr()
                                               .c_str());
                                    assert(0 &&
                                           "not int type and not float type");
                                }
                            }
                        } else {
                            assert(0 && "gep op1 type is not ptr");
                        }
                    }
                    // old2new[ins]->getOp()[1]->getSymbolEntry()->setType(TypeSystem::intType);
                    // std::cout<<"old2new[ins]->getOp()[i] =
                    // "<<old2new[ins]->getOp()[i]->getType()->toStr()<<"\n";
                    // std::cout<<"old2new[ins]->getOp()[i] =
                    // "<<old2new[ins]->getOp()[i]->toStr()<<"\n";
                    old2new[ope]->getOp()[0]->addUse(old2new[ins]);
                } else {
                    // print("branch else");
                    old2new[ins]->setOperand(i, ope->getOp()[0]);
                    ope->getOp()[0]->addUse(old2new[ins]);
                }
                // std::cout<<"small loop end\n";
            }
        }
        // print("after transfer body_bb for exchange");
        for (Instruction *ins = new_bb->begin(); ins != new_bb->end();
             ins = ins->getNext()) {
            // if(ins->isCall())
            // {
            //     std::cout<<"call dst = "<<ins->getDef()->getLabel()<<" call
            //     dst =
            //     "<<dynamic_cast<CallInstruction*>(ins)->dst->getLabel()<<"\n";
            // }
        }
    }
// for(Instruction* ins : sitofps)
// {
//     printf("one sitofps ins op0 = %d, op1 = %d\n",
//     ins->getOp()[0]->getLabel(), ins->getOp()[1]->getLabel());
// }
#ifdef DEBUG0429
    print("after while(start < end)");
#endif
    // std::cout<<"stage4\n";
    old2new[iter] = new BinaryInstruction(
        BinaryInstruction::ADD, iter->getDef(),
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, start)),
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, accumulator-1)),
        new_bb);

    old2new.clear();
    std::unordered_map<Instruction *, Instruction *> old2new_;
#ifdef DEBUG0429
#endif
    Instruction *outbb_phi = nullptr;
    Operand *outbb_phi_dst;
    for (Instruction *ins = outBB->begin(); ins != outBB->end();
         ins = ins->getNext()) {
        // std::cout<<"ins = "<<ins<<"\n";
        // printf("is sitofp(outbb) = %d\n", ins->isSitofp());
        if (ins->isPhi()) {
            outbb_phi = ins;
            outbb_phi_dst = ins->getDef();
            print("outbb_phi_dst = " << outbb_phi_dst->toStr() << "");
        }
        // std::cout<<"ins.type = "<<ins->getInstName()<<"\n";
        // std::cout<<"ins.ops = "<<ins->getOp().size()<<"\n";
        for (unsigned i = 0; i < ins->getOp().size(); i++) {
            // std::cout<<"breakpoint0\n";
            if (ins->getOp()[i] == nullptr) {
                continue;
            }
            Instruction *ope = ins->getOp()[i]->getDef();
            if (ope == nullptr) {
                // std::cout<<"breakpoint1\n";
#ifdef DEBUG0429_terminal
                print("ope == nullptr");
#endif
                continue;
            }
            if (old2new_.find(ope) != old2new_.end()) {
                // std::cout<<"breakpoint2\n";
                ins->setOperand(i, old2new_[ope]->getOp()[0]);
            } else if (ope->isPhi()) {
                // std::cout<<"breakpoint3\n";
                ins->setOperand(
                    i, new Operand(new ConstantSymbolEntry(
                           TypeSystem::constIntType, start + accumulator)));
            }
            // std::cout<<"breakpoint4\n";
        }
    }

    // printf("bodybb = %d\n", body_bb->getNo());
    // printf("condbb label = %d\n", condbb->getNo());
    // printf("bodybb label = %d\n", bodybb->getNo());
    // wzr!!: 这里下面的getPredBlock()直接取[0]是有可能导致问题的
    bool link = false;
    for (auto succ : cond_bb->getSuccBlock()) {
        // std::cout<<"succ "<<succ->getNo()<<" body "<<body_bb->getNo()<<"\n";
        if (succ == body_bb)
            link = true;
    }

    BasicBlock *mtempbb = condbb->getPredBlock()[0];
    BasicBlock *mtembbb_prev = mtempbb->getPredBlock()[0];
    // Operand* mop2 = body_bb->begin()->getNext()->getOp()[2];
    Operand *mop2 = nullptr;
    for (Instruction *mins = body_bb->begin(); mins != body_bb->end();
         mins = mins->getNext()) {
        // printf("!!!mins = %s\n", mins->getInstName().c_str());
        // printf("mins = %s\n", mins->getInstName().c_str());
        if (mins->isGep() && !mins->getNext()->isGep()) {
            if (mins->getPrev()->isGep()) {
                // printf("mop2 branch1");
                mop2 = mins->getPrev()->getOp()[2];
            } else {
                // printf("mop2 branch2");
                mop2 = mins->getOp()[2];
            }
            // printf("in!!");
            // mtempbb->insertBefore(i_str, mins);
            break;
        }
    }

    assert(mop2 != nullptr);
    // printf("mop2 = %d\n", mop2->getLabel());
    // printf("traverse end_bb, find i's addr");
    LoadInstruction *i_load = nullptr;
    Instruction *i_binary = nullptr;
    // printf("end_bb no = %d\n", end_bb->getNo());
    for (Instruction *ins = end_bb->begin(); ins != end_bb->end();
         ins = ins->getNext()) {
        // printf("ins = %d\n", ins);
        // printf("ins = %s\n", ins->getInstName().c_str());
        // printf("op size = %d\n", ins->getOp().size());
        // for(int i = 0; i < ins->getOp().size(); i ++)
        // {
        //     if(ins->getOp()[i] == nullptr)
        //     {
        //         // printf("op%d == nullptr, next\n", i);
        //         continue;
        //     }
        //     // assert(0 && "ins->getOp()[i] == nullptr");
        //     // printf("ins op%d = %d\n", i, ins->getOp()[i]->getLabel());
        // }
        // printf("after print ops");
        if (ins->isBin()) {
            Operand *op0 = ins->getOp()[0];
            // printf("binary op0 = %d\n", op0->getLabel());
            // assert(op0->getDst() != nullptr);
            // printf("op0 getDst = %d\n", op0->getDst()->getLabel());
            if (op0->getDst() == nullptr) {
                break;
            }
            StoreInstruction *i_str =
                new StoreInstruction(op0->getDst(), mop2, nullptr);
            // mtempbb->insertFront(i_str);
            // printf("mtempbb no = %d\n", mtempbb->getNo());
            for (Instruction *mins = mtempbb->begin(); mins != mtempbb->end();
                 mins = mins->getNext()) {
                // printf("!!!mins = %s\n", mins->getInstName().c_str());
                if (!mins->isPhi()) {
                    // printf("in!!");
                    mtempbb->insertBefore(i_str, mins);
                    break;
                }
            }

            i_load = new LoadInstruction(op0, op0->getDst(), nullptr);

            ins->setOperand(1, op0);
        }
        if (ins->isBin()) {
            i_binary = ins;
        }
    }
    print("stage5");
    // 把i_load插入end_bb中binary指令前
    if (i_binary != nullptr) {
        end_bb->insertBefore(i_load, i_binary);
    }

    // 向end_bb里插入从i_addr中load i值的语句

    // std::cout<<"body_bb->getSuccBlock().size()1 =
    // "<<body_bb->getSuccBlock().size()<<"\n";
    //  Instruction* binaryins = new BinaryInstruction(BinaryInstruction::ADD,
    //  outbb_phi_dst, iter->getDef(),
    //                                     new Operand(new
    //                                     ConstantSymbolEntry(TypeSystem::intType,
    //                                     0)));
    //  outBB->insertFront(binaryins);

    // std::cout<<outbb_phi->getParent()->getNo()<<"
    // "<<outbb_phi->getDef()->toStr()<<" ???????\n";
    if (outbb_phi != nullptr && outbb_phi->getParent() == outBB)
        outBB->remove(outbb_phi);

    // clear

    // cond_bb->clearInstructions();
    // body_bb->clearInstructions();

    // return;
    //  std::cout<<"new_bb: "<<new_bb->getNo()<<"\n";
    //  std::cout<<"end_bb: "<<end_bb->getNo()<<"\n";
    if (outBB->existsInPred(body_bb)) {
        outBB->removePred(body_bb);
    }
    if (body_bb->existsInSucc(outBB)) {
        body_bb->removeSucc(outBB);
    }
    UncondBrInstruction *uncondbr = new UncondBrInstruction(outBB, new_bb);
    new_bb->addSucc(outBB);
    outBB->addPred(new_bb);
    if (find_inst(Instruction::CMP, cond_bb)) {
        auto cmp = dynamic_cast<CmpInstruction *>(
            find_inst(Instruction::CMP, cond_bb));
        cmp->getParent()->remove(cmp);
    }
    if (find_inst(Instruction::COND, cond_bb)) {
        auto cond = dynamic_cast<CondBrInstruction *>(
            find_inst(Instruction::COND, cond_bb));
        cond->getParent()->remove(cond);
    }
    if (find_inst(Instruction::UNCOND, cond_bb)) {
        auto uncond = dynamic_cast<UncondBrInstruction *>(
            find_inst(Instruction::UNCOND, cond_bb));
        uncond->setBranch(new_bb);
    } else
        new UncondBrInstruction(new_bb, cond_bb);
    if (body_bb->existsInPred(cond_bb)) {
        body_bb->removePred(cond_bb);
    }
    if (cond_bb->existsInSucc(body_bb)) {
        cond_bb->removeSucc(body_bb);
    }
    if (!new_bb->existsInPred(cond_bb))
        new_bb->addPred(cond_bb);
    if (!cond_bb->existsInSucc(new_bb))
        cond_bb->addSucc(new_bb);
    //     for (BasicBlock *pred_cond : cond_bb->getPredBlock()) {
    //         print(">>>>>>>>>>>>>>>>>>>>> cond_bb: " << cond_bb->getNo());
    //         print(">>>>>>>>>>>>>>>>>>>>> pred_cond: " << pred_cond->getNo());
    //         if (pred_cond == body_bb) {
    // #ifdef DEBUG0429_terminal
    //             print("pred_cond == body_bb");
    // #endif
    //             continue;
    //         }
    //         Instruction *b = pred_cond->getBrorRet();
    //         if (link) {
    //             if (b->getInstName() == "COND") {
    //                 CondBrInstruction *condbr =
    //                     dynamic_cast<CondBrInstruction *>(b);
    //                 if (condbr->getTrueBranch() == cond_bb) {
    //                     condbr->setTrueBranch(new_bb);
    //                     pred_cond->addSucc(new_bb);
    //                     pred_cond->removeSucc(cond_bb);
    //                     // new_bb->removePred(pred_cond);
    //                     new_bb->addPred(pred_cond);
    //                 } else if (condbr->getFalseBranch() == cond_bb) {
    //                     condbr->setFalseBranch(new_bb);
    //                     pred_cond->addSucc(new_bb);
    //                     pred_cond->removeSucc(cond_bb);

    //                     new_bb->addPred(pred_cond);
    //                 }
    //             } else if (b->getInstName() == "UNCOND") {
    //                 UncondBrInstruction *uncondbr =
    //                     dynamic_cast<UncondBrInstruction *>(b);
    //                 if (uncondbr->getBranch() == cond_bb) {
    //                     uncondbr->setBranch(new_bb);
    //                     pred_cond->addSucc(new_bb);
    //                     pred_cond->removeSucc(cond_bb);
    //                     new_bb->addPred(pred_cond);
    //                 }
    //             }
    //         } else {
    //             BasicBlock *invar_bb;
    //             for (auto pred : body_bb->getPredBlock()) {
    //                 if (pred != body_bb) {
    //                     invar_bb = pred;
    //                     break;
    //                 }
    //             }
    //             if (invar_bb == nullptr)
    //                 assert(0);

    //             auto uncond_ = (UncondBrInstruction *)(find_inst(
    //                 Instruction::UNCOND, invar_bb));
    //             uncond_->setBranch(new_bb);

    //             invar_bb->addSucc(new_bb);
    //             invar_bb->removeSucc(body_bb);
    //             body_bb->removePred(invar_bb);
    //             new_bb->addPred(invar_bb);
    //             if (b->getInstName() == "COND") {
    //                 CondBrInstruction *condbr =
    //                     dynamic_cast<CondBrInstruction *>(b);
    //                 if (condbr->getTrueBranch() == cond_bb) {
    //                     condbr->setTrueBranch(invar_bb);
    //                     pred_cond->addSucc(invar_bb);
    //                     pred_cond->removeSucc(cond_bb);
    //                     // invar_bb->removePred(pred_cond);
    //                     invar_bb->addPred(pred_cond);
    //                 } else if (condbr->getFalseBranch() == cond_bb) {
    //                     condbr->setFalseBranch(invar_bb);
    //                     pred_cond->addSucc(invar_bb);
    //                     pred_cond->removeSucc(cond_bb);

    //                     invar_bb->addPred(pred_cond);
    //                 }
    //             } else if (b->getInstName() == "UNCOND") {
    //                 UncondBrInstruction *uncondbr =
    //                     dynamic_cast<UncondBrInstruction *>(b);
    //                 if (uncondbr->getBranch() == cond_bb) {
    //                     uncondbr->setBranch(invar_bb);
    //                     pred_cond->addSucc(invar_bb);
    //                     pred_cond->removeSucc(cond_bb);
    //                     invar_bb->addPred(pred_cond);
    //                 }
    //             }
    //         }
    //     }

    print("stage6");

    // std::cout<<"cond_bb = "<<cond_bb->getNo()<<"\n";
    // if (cond_bb != nullptr)
    //     cond_bb->deleteAllRelation();
    // cond_bb->getParent()->remove(cond_bb);
    if (body_bb != nullptr)
        body_bb->deleteAllRelation();
    body_bb->getParent()->remove(body_bb);
    print("after delete body_bb = " << body_bb->getNo());
    std::cout << "yes unfold done.\n";
    print("========= end =========");
}
