#include "DeadCodeEliminate.h"
#include <unordered_set>
#include <vector>

#define DEBUG_SWITCH_eli 0
#if DEBUG_SWITCH_eli
#define print(str) std::cout << str << "\n";
#else
#define print(str) //
#endif

using namespace std;

void printFrontBB(BasicBlock *bb) {
    std::set<BasicBlock *>::iterator it;
    for (it = bb->frontierBlock->begin(); it != bb->frontierBlock->end();
         ++it) {
        print((*it)->getNo());
    }
}
void DeadCodeEliminate::clearNoPred() {
    for (auto func : unit->get_functions()) {
        while (func->clearUselessBB())
            ;
    }
}

void DeadCodeEliminate::executeFront() {
    for (auto func : unit->get_functions()) {
        if (dynamic_cast<IdentifierSymbolEntry *>(func->getSymPtr())
                ->getName() == "main")
            continue;
        deleteDeadStore(func);
    }
}

void DeadCodeEliminate::execute() {
    print("============ DCE start ============");
    deadBranchDetection();
    clearNoPred();
    // for (auto func : unit->get_functions()) {
    //     print("## DeadCodeEliminate "<<func->getSymPtr()->toStr());
    //     bool again = true;
    //     // while (again) {
    //         // adjustBlock(func);
    //         mark(func);
    //         print("marked");
    //         again = remove(func);
    //         print("removed");
    //         // adjustBlock(func);
    //     // }
    // }

    // todo: dce
    // deleteDeadFunc();
    // detectNotSideEffectFunc();
    // for (auto func : unit->get_functions()) {
    //     if (func != unit->getMainFunction())
    //         deleteDeadRet(func);
    // }
    print("============ DCE deleteDeadInst ============");
    for (auto func : unit->get_functions()) {
        deleteDeadInst(func);
        // deleteDeadStore(func);
        // deleteDeadInst(func);
    }
    print("============ DCE finish ============");
}

bool delBranchRelation(Function *func, BasicBlock *bb, BasicBlock *branch) {
    auto succs = bb->getSuccBlock();
    if (std::find(succs.begin(), succs.end(), branch) != succs.end())
        bb->removeSucc(branch);
    auto preds = branch->getPredBlock();
    if (std::find(preds.begin(), preds.end(), bb) != preds.end())
        branch->removePred(bb);

    if (branch->predEmpty() ||
        (branch->existsInPred(branch) && branch->getNumOfPred() == 1)) {
        branch->deleteAllRelation();
        func->remove(branch);
    }
    return 1;
}

void DeadCodeEliminate::deadBranchDetection() {
    std::set<Instruction *> delete_list = {};
    for (auto func : unit->get_functions()) {
        // auto blockList = func->getBlockList();
        // for (int i = 0; i < blockList.size(); i++) {
        // auto bb = blockList[i];
        for (auto bb : func->getBlockList()) {
            print(" B" << bb->getNo());
            // printFrontBB(bb);
            if (bb->predEmpty() && bb != *func->begin() ||
                (bb->existsInPred(bb) && bb->getNumOfPred() == 1)) {
                print("== B" << bb->getNo());
                bb->deleteAllRelation();
                func->remove(bb);
                continue;
            }
            delete_list.clear();
            auto head = bb->end();
            for (auto instr = head->getNext(); instr != head;
                 instr = instr->getNext()) {
                if (instr->isCond()) {
                    auto cond_instr = dynamic_cast<CondBrInstruction *>(instr);

                    auto giveV = instr->getOp()[0]->getDef();
                    auto se1 = giveV->getOp()[1]->getSymbolEntry();
                    auto se2 = giveV->getOp()[2]->getSymbolEntry();
                    if (giveV->isCmp() && se1->isConstant() &&
                        se2->isConstant()) {
                        BasicBlock *true_bb = cond_instr->getTrueBranch();
                        BasicBlock *false_bb = cond_instr->getFalseBranch();
                        double value1 = dynamic_cast<ConstantSymbolEntry *>(se1)
                                            ->getValue();
                        double value2 = dynamic_cast<ConstantSymbolEntry *>(se2)
                                            ->getValue();
                        // print(value1 << " "<<giveV->getOpcode()<<" " <<
                        // value2);
                        bool res =
                            getRelationRes(giveV->getOpcode(), value1, value2);
                        /** a || b
                         * TrueBranch相同
                         */
                        // 那看来还是要输出的时候删除
                        if (res == 0) {
                            auto uncond = new UncondBrInstruction(false_bb);
                            print("del true_bb = " << true_bb->getNo());
                            bb->insertAfter(uncond, instr);
                            // for (auto succ : true_bb->getSuccBlock()) {
                            //     print("del true_bb succ= " << succ->getNo());
                            //     if (succ == true_bb->getBranchEnd())
                            //         break;
                            //     succ->deleteAllRelation();
                            //     func->remove(succ);
                            // }

                            // auto cond = true_bb->getBranchBB();
                            delBranchRelation(func, bb, true_bb);
                            // BasicBlock *true_bb_t = nullptr;
                            // BasicBlock *false_bb_t = nullptr;
                            // bool update = false;
                            // while(cond!=nullptr){
                            //     true_bb_t = cond->getTrueBranch();
                            //     false_bb_t = cond->getFalseBranch();
                            // }
                            // do {
                            //     if (cond != nullptr) {
                            //         print("=============");
                            //         update = true;

                            //     }
                            //     if()
                            // } while (update);

                        } else {
                            auto uncond = new UncondBrInstruction(true_bb);
                            print("del false_bb = " << false_bb->getNo());
                            bb->insertAfter(uncond, instr);
                            // for (auto succ : false_bb->getSuccBlock()) {
                            //     if (succ == false_bb->getBranchEnd())
                            //         break;
                            //     succ->deleteAllRelation();
                            //     func->remove(succ);
                            // }

                            delBranchRelation(func, bb, false_bb);

                            // auto succs = bb->getSuccBlock();
                            // if (std::find(succs.begin(), succs.end(),
                            //               true_bb) != succs.end())
                            //     bb->removeSucc(false_bb);
                            // auto preds = false_bb->getPredBlock();
                            // if (std::find(preds.begin(), preds.end(), bb) !=
                            //     preds.end())
                            //     false_bb->removePred(bb);

                            // if (false_bb->predEmpty() ||
                            //     (false_bb->existsInPred(false_bb) &&
                            //      false_bb->getNumOfPred() == 1)) {
                            //     false_bb->deleteAllRelation();
                            //     func->remove(false_bb);
                            // }
                        }
                        delete_list.insert(instr);
                        delete_list.insert(giveV);
                    }
                }
            }
            for (auto instr : delete_list) {
                print(instr->getInstName());
                bb->remove(instr);
            }
        }
    }
}

bool DeadCodeEliminate::getRelationRes(unsigned opcode, double value1,
                                       double value2) {
    bool res;
    switch (opcode) {
    case CmpInstruction::E:
        print(value1 << " E " << value2);
        res = (value1 == value2);
        break;
    case CmpInstruction::NE:
        print(value1 << " NE " << value2);
        res = (value1 != value2);
        break;
    case CmpInstruction::L:
        print(value1 << " L " << value2);
        res = (value1 < value2);
        break;
    case CmpInstruction::LE:
        print(value1 << " LE " << value2);
        res = (value1 <= value2);
        break;
    case CmpInstruction::G:
        print(value1 << " G " << value2);
        res = (value1 > value2);
        break;
    case CmpInstruction::GE:
        print(value1 << " GE " << value2);
        res = (value1 >= value2);
        break;
    default:
        print("sth wrong");
        break;
    }
    return res;
}

void DeadCodeEliminate::deleteDeadFunc() {
    // func_counter = 0;
    // for (auto func : unit->get_functions()) {
    //     if (func->getBlockList().empty())
    //         continue;

    //     // if (func->getUseList().empty() && func != unit->getMainFunction())
    //     {
    //     //   func_counter++;
    //     //   iter = funcs.erase(iter);
    //     // }
    // }
}

void DeadCodeEliminate::detectNotSideEffectFunc() {
    for (auto func : unit->get_functions()) {
        auto func_se = dynamic_cast<IdentifierSymbolEntry *>(func->getSymPtr());
        print(func_se->getName());
        print(dynamic_cast<FunctionType *>(func_se->getType())
                  ->getRetType()
                  ->toStr());
        if (!(dynamic_cast<FunctionType *>(func_se->getType())
                  ->getRetType()
                  ->toStr() == "void"))
            continue;
        print("in pass");
        // if
        // (dynamic_cast<IdentifierSymbolEntry*>(func->getSymPtr())->getName()
        // == "main")
        //   continue;
        bool side_effect = false;
        for (auto bb : func->getBlockList()) {
            auto head = bb->end();
            for (auto instr = head->getNext(); instr != head;
                 instr = instr->getNext()) {
                if (instr->isStore()) {
                    auto se = instr->getDef()->getOpdSe();
                    if (dynamic_cast<IdentifierSymbolEntry *>(se)->isGlobal()) {
                        side_effect = true;
                        continue;
                        // break;
                    } else {
                        if (instr->getDef()->usersNum() <= 1) {
                            // print(instr->getDef()->toStr());
                            bb->remove(instr);
                        }
                    }

                } else if (instr->isCall()) {
                    side_effect = true;
                    break;
                }
            }
        }
        if (side_effect == false) {
            print("side_effect == false");
            // func->getBlockList().clear();
            // BasicBlock *bb = new BasicBlock(func);
            // new RetInstruction(nullptr, bb);
        }
        // notSideEffectFunc.insert(func);
    }
    print("======== notSideEffectFunc ==========");
}

bool DeadCodeEliminate::isSideEffect(Instruction *instr) {
    if (instr->isRet() || instr->isUncond() || instr->isCond() ||
        instr->isCmp() ||
        // instr->isStore() ||
        instr->isCall() // todo:call
    ) {
        return true;
    } else if (instr->isStore()) {
        if (instr->getOpse().size() == 0)
            return false;
        auto se = dynamic_cast<StoreInstruction *>(instr)->getIdentse();
        if (se == nullptr)
            return true; // 数组移位赋值
        print("in " << se->toStr());
        if (se->isParam() || se->isGlobal() ||
            se->getType()->isArray()) { // 还是先避开数组赋值
            return true;
        }
        if (worklist_op.find(instr->getDef()->toStr()) != worklist_op.end()) {
            return true;
        }
        for (auto op : instr->getUse()) {
            if (worklist_op.find(op->toStr()) != worklist_op.end()) {
                return true;
            }
        }
    } else if (instr->isPhi()) {
        if (instr->getOpse().size() == 0)
            return false;
        auto se = dynamic_cast<PhiInstruction *>(instr)->getIdentse();
        print("in " << se->toStr());
        if (se->isParam() || se->isGlobal() ||
            se->getType()->isArray()) { // 还是先避开数组赋值
            return true;
        }
    }
    return false;
}

bool DeadCodeEliminate::isSideEffect_main(Instruction *instr) {
    if (instr->isRet() || instr->isUncond() || instr->isCond() ||
        instr->isCmp() || instr->isStore() || instr->isPhi() ||
        instr->isCall() // todo:call
    ) {
        return true;
    }
    else{
        if (worklist_op.find(instr->getDef()->toStr()) != worklist_op.end()) {
            return true;
        }
    }
    return false;
}

void DeadCodeEliminate::markUse_print(Instruction *instr, int level) {
    for (int i = 0; i < level; i++)
        std::cout << " ";
    std::cout << "  | " << instr->getInstName() << "\n";
    if (worklist.find(instr) != worklist.end())
        return;
    worklist.insert(instr);
    for (auto op : instr->getUse()) {
        worklist_op.insert(op->toStr());
        if (op->getDef())
            markUse_print(op->getDef(), level + 1);
    }
}

void DeadCodeEliminate::markUse(Instruction *instr) {
    // if (instr->getDef())
    //     print("  | " << instr->getInstName() << " "
    //                  << instr->getOp()[0]->toStr());
    // print(&instr);

    // if (worklist.find(instr) != worklist.end())
    //     return;
    // worklist.insert(instr);

    bool needMark = 0;
    if (worklist.find(instr) == worklist.end()) {
        needMark = 1;
        worklist.insert(instr);
    }
    // if (instr->getDef() && !needMark&&
    //     worklist_op.find(instr->getDef()->toStr()) != worklist_op.end()) {
    //     needMark = 1;
    // }
    if (!needMark)
        return;
    if (instr->isStore()) {
        auto def = instr->getDef();
        print(def->toStr());
        // worklist_op.insert(def->toStr());
        if (def->getDef())
            markUse(def->getDef());
        for (auto op : instr->getUse()) {
            if (op->getDef()) {
                if (worklist_op.find(op->toStr()) == worklist_op.end())
                    worklist_op.insert(op->toStr());
                markUse(op->getDef());
            }
        }
    } else if (instr->isPhi()) {
        PhiInstruction *phi = dynamic_cast<PhiInstruction *>(instr);
        for (auto op : phi->getUse()) {
            if (worklist_op.find(op->toStr()) == worklist_op.end())
                worklist_op.insert(op->toStr());
            // print("  ===== " << op->toStr());
            if (op->getDef())
                markUse(op->getDef());
        }
    } else {

        for (auto op : instr->getUse()) {

            if (worklist_op.find(op->toStr()) == worklist_op.end())
                worklist_op.insert(op->toStr());
            if (op->getDef()) {
                // if (instr->getDef()&&instr->getDef()->toStr() == "%t117") {
                //     print("here");
                //     std::cout << "  ===== " << op->toStr() << "\n";
                // std::cout << op->getDef()->getDef()->getUse().size() << "\n";
                // std::cout << op->getDef()->getInstName()<<"\n";
                // }
                print(op->getDef()->getInstName());
                markUse(op->getDef());
            }
        }
    }
}

void DeadCodeEliminate::deleteDeadInstFront(Function *func) {
    worklist.clear();
    worklist_op.clear();
    for (auto bb : func->getBlockList()) {
        print("  B" << bb->getNo());
        if (dynamic_cast<IdentifierSymbolEntry *>(func->getSymPtr())
                ->getName() == "main") {
            for (auto instr = bb->begin(); instr != bb->end();
                 instr = instr->getNext()) {
                print(" + " << instr->getInstName());
                if (isSideEffect_main(instr)) {
                    markUse(instr);
                }
            }
        } else {
            for (auto instr = bb->begin(); instr != bb->end();
                 instr = instr->getNext()) {
                // std::cout << " + " << instr->getInstName() << "\n";
                print(" + " << instr->getInstName());
                if (isSideEffect(instr)) {
                    markUse(instr);
                }
            }
        }
    }
    print("========== finish mark ===========");
    for (auto bb : func->getBlockList()) {
        std::unordered_set<Instruction *> removelist;
        for (auto instr = bb->begin(); instr != bb->end();
             instr = instr->getNext()) {
            if (worklist.find(instr) == worklist.end() &&
                worklist_op.find(instr->getDef()->toStr()) ==
                    worklist_op.end()) {
                removelist.insert(instr);
            }
        }
        // std::cout << " remove size: " << removelist.size() << "\n";
        for (auto instr : removelist) {
            bb->remove(instr);
            inst_counter++;
        }
    }
}

void DeadCodeEliminate::deleteDeadInst(Function *func) {
    worklist.clear();
    worklist_op.clear();
    for (auto bb : func->getBlockList()) {
        print("  B" << bb->getNo());
        if (dynamic_cast<IdentifierSymbolEntry *>(func->getSymPtr())
                ->getName() == "main") {
            for (auto instr = bb->begin(); instr != bb->end();
                 instr = instr->getNext()) {
                print(" + " << instr->getInstName());
                if (isSideEffect_main(instr)) {
                    markUse(instr);
                }
            }
        } else {
            for (auto instr = bb->begin(); instr != bb->end();
                 instr = instr->getNext()) {
                // std::cout << " + " << instr->getInstName() << "\n";
                print(" + " << instr->getInstName());
                if (isSideEffect(instr)) {
                    markUse(instr);
                }
            }
        }
    }
    print("========== finish mark ===========");
    for (auto bb : func->getBlockList()) {
        std::unordered_set<Instruction *> removelist;
        for (auto instr = bb->begin(); instr != bb->end();
             instr = instr->getNext()) {
            if (worklist.find(instr) == worklist.end() &&
                // 为了解决phi的多def
                worklist_op.find(instr->getDef()->toStr()) ==
                    worklist_op.end()) {
                removelist.insert(instr);
            }
        }
        // std::cout << " remove size: " << removelist.size() << "\n";
        for (auto instr : removelist) {
            // if (instr->isCond()) {
            //     new UncondBrInstruction(
            //         dynamic_cast<CondBrInstruction
            //         *>(instr)->getFalseBranch(), bb);
            //     bb->remove(instr);
            //     continue;
            // }
            bb->remove(instr);
            inst_counter++;
        }
        // if(bb->empty())
    }
}

void DeadCodeEliminate::deleteDeadStore(Function *func) {
    std::unordered_set<Instruction *> storelist;
    for (auto bb : func->getBlockList()) {
        print("  B" << bb->getNo());
        for (auto instr = bb->begin(); instr != bb->end();
             instr = instr->getNext()) {
            if (instr->isStore()) {
                auto se = dynamic_cast<StoreInstruction *>(instr)->getIdentse();
                if (se == nullptr || se->isParam() || se->isGlobal() ||
                    se->getType()->isArray()) { // 还是先避开数组赋值
                    continue;
                }
                auto def = instr->getDef();
                if (def->getUse().size() <= 1) {
                    storelist.insert(instr);
                    // std::cout<<def->getDef()->getInstName()<<"\n";
                    if (!def->getDef()->isAlloc())
                        assert(0);
                    storelist.insert(def->getDef()); // Alloc
                }
            }
        }
        for (auto instr : storelist) {
            bb->remove(instr);
        }
    }
    print("========== finish delStore ===========");
}

void DeadCodeEliminate::deleteDeadRet(Function *func) {
    // bool flag = true;
    // for (auto use : func->getUseList()) {
    //     if (!use.val_->getUseList().empty())
    //         flag = false;
    // }
    // if (flag) {
    //     for (auto bb : func->getBlockList()) {
    //         for (auto instr : bb->getInstructions()) {
    //             if (instr->isRet() && instr->getNumOperand() == 1) {
    //                 instr->removeUseOfOps();
    //                 instr->setOperand(0, ConstantInt::get(0, unit));
    //             }
    //         }
    //     }
    // }
}