#include "Function.h"
#include "Type.h"
#include "Unit.h"
#include <list>

extern FILE *yyout;

bool Function::clearUselessBB() {
    bool update = false;
    for (auto bb : block_list) {
        if (bb->predEmpty() && bb != block_list[0] ||
            (bb->existsInPred(bb) && bb->getNumOfPred() == 1)) {
            bb->deleteAllRelation();
            remove(bb);
            update=true;
        }
    }
    return update;
}

Function::Function(Unit *u, SymbolEntry *s) {
    u->insertFunc(this);
    entry = new BasicBlock(this);
    ((IdentifierSymbolEntry *)s)->setFunction(this); //!
    sym_ptr = s;
    parent = u;
}

Function::~Function() {
    // auto delete_list = block_list;
    // for (auto &i : delete_list)
    //     delete i;
    // parent->removeFunc(this);
}

// remove the basicblock bb from its block_list.
void Function::remove(BasicBlock *bb) {
    auto iter = std::find(block_list.begin(), block_list.end(), bb);
    if(iter!=block_list.end())
    block_list.erase(iter);
}

void Function::output() const {
    FunctionType *funcType = dynamic_cast<FunctionType *>(sym_ptr->getType());
    Type *retType = funcType->getRetType();
    std::vector<SymbolEntry *> paramsSe = funcType->getParamsSe();
    if (!paramsSe.size())
        fprintf(yyout, "define %s %s() {\n", retType->toStr().c_str(),
                sym_ptr->toStr().c_str());
    else {
        fprintf(yyout, "define %s %s(", retType->toStr().c_str(),
                sym_ptr->toStr().c_str());
        for (long unsigned int i = 0; i < paramsSe.size(); i++) {
            if (i)
                fprintf(yyout, ", ");
            fprintf(yyout, "%s %s", paramsSe[i]->getType()->toStr().c_str(),
                    paramsSe[i]->toStr().c_str());
        }
        fprintf(yyout, ") {\n");
    }
    std::set<BasicBlock *> v;
    std::list<BasicBlock *> q;
    q.push_back(entry);
    v.insert(entry);

    while (!q.empty()) {
        auto bb = q.front();
        q.pop_front();
        bb->output();
        for (auto succ = bb->succ_begin(); succ != bb->succ_end(); succ++) {
            if (v.find(*succ) == v.end()) {
                v.insert(*succ);
                q.push_back(*succ);
            }
        }
    }
    fprintf(yyout, "}\n");
}

void Function::genMachineCode(AsmBuilder *builder) {
    // std::cout<<sym_ptr->toStr()<<" ========== func genmachinecode\n";
    auto cur_unit = builder->getUnit();
    auto cur_func = new MachineFunction(cur_unit, this->sym_ptr);
    builder->setFunction(cur_func);
    BasicBlock* tmp;
    // for(auto bb:block_list){
    //     if(bb->getNo()==350){
    //         std::cout<<"found\n";
    //         tmp=bb;
    //     }
    // }
    // bool ok=0;
    for (auto block : block_list) {
        // if(ok)std::cout<<this->map2[tmp]<<"\n";
        // std::cout<<"before block genmachinecode\n";
        // if(block->getNo()==350)ok=1;
        block->genMachineCode(builder);
        this->map2[block] = builder->getBlock();
        // std::cout<<block->getNo()<<"\t"<<block;
        // std::cout<<"\t"<<this->map2[block]<<"\n";
    }
    // Add pred and succ for every block
    // std::cout<<this->map2[tmp]<<" breakpoint1\n";
    for (auto block : block_list) {
        // std::cout<<"block: "<<block->getNo();
        // std::cout<<" "<<this->map2[tmp]<<"\n";
        auto mblock = this->map2[block];
        for (auto pred = block->pred_begin(); pred != block->pred_end(); pred++){
        // if(block->getNo()==523||block->getNo()==350)std::cout<<"=========="<<block->getNo()<<" pred: "<<tmp<<"\n";
            mblock->addPred(this->map2[*pred]);
        }
        for (auto succ = block->succ_begin(); succ != block->succ_end(); succ++){
        // if(block->getNo()==523||block->getNo()==350)std::cout<<"=========="<<block->getNo()<<" succ: "<<(*succ)->getNo()<<" "<<(*succ)<<"\n";
            mblock->addSucc(this->map2[*succ]);}
    }
    // getchar();
    cur_func->setEntry(map2[entry]);
    cur_unit->InsertFunc(cur_func);
    // std::cout<<"func genmachinecode end\n";
    //  0413m
    //  在所有机器代码都生成完毕的情况下进行phi指令的机器代码生成
    //  fix: 有问题
    //  for(auto block : block_list)
    //  {
    //      for(auto i = block->begin(); i != block->end(); i = i->getNext()){
    //          if(!i->isPhi()){
    //              break;
    //          }
    //          // std::cout<<block->getNo()<<"生成phi指令: \n";
    //          auto phi = dynamic_cast<PhiInstruction*>(i);
    //          phi->genMachineCodeAfter(builder);
    //          // std::cout<<"生成结束 \n";
    //      }
    //  }
}

int Function::getEssential() {
    if (essential != -1)
        return essential;
    FunctionType *type = (FunctionType *)(sym_ptr->getType());
    auto paramsType = type->getParamsType();
    for (auto type : paramsType)
        if (type->isArray()) {
            essential = 1;
            return essential;
        }
    for (auto block : block_list) {
        for (auto it = block->begin(); it != block->end(); it = it->getNext()) {
            if (it->isCall()) {
                IdentifierSymbolEntry *funcSE =
                    (IdentifierSymbolEntry *)(((CallInstruction *)it)
                                                  ->getFunc_se());
                if (funcSE->isSysy() ||
                    funcSE->getName() == "llvm.memset.p0.i32") {
                    essential = 1;
                    return essential;
                } else {
                    auto func = funcSE->getFunction();
                    if (func == this)
                        continue;
                    if (func->getEssential() == 1) {
                        essential = 1;
                        return essential;
                    }
                }
            } else {
                auto def = it->getDef();
                if (def && def->getEntry()->isVariable()) {
                    auto se = (IdentifierSymbolEntry *)(def->getEntry());
                    if (se->isGlobal()) {
                        essential = 1;
                        return essential;
                    }
                }
                auto uses = it->getUse();
                for (auto use : uses) {
                    if (use && use->getEntry()->isVariable()) {
                        auto se = (IdentifierSymbolEntry *)(use->getEntry());
                        if (se->isGlobal()) {
                            essential = 1;
                            return essential;
                        }
                    }
                }
            }
        }
    }
    essential = 0;
    return essential;
}

void Function::addPred(Instruction *in) {
    // assert(in->isCall());
    // auto func = in->getParent()->getParent();
    // if (func == this)
    //     recur = true;
    // if (preds.count(func))
    //     preds[func].push_back(in);
    // else
    //     preds[func] = {in};
}

void Function::removePred(Instruction *in) {
    assert(in->isCall());
    auto func = in->getParent()->getParent();
    auto it = find(preds[func].begin(), preds[func].end(), in);
    assert(it != preds[func].end());
    preds[func].erase(it);
}

BasicBlock *Function::getMarkBranch(BasicBlock *block) {
    // std::set<BasicBlock*> blocks;
    // while (true) {
    //     auto order = idoms[block->order];
    //     block = preOrder2dom[order]->block;
    //     if (blocks.count(block))
    //         return nullptr;
    //     blocks.insert(block);
    //     if (block->isMark())
    //         return block;
    // }
    return nullptr;
}