#include "LoopDetection.h"

using blockSet = std::unordered_set<BasicBlock *>;

#define DEBUG_SWITCH_ld 0
#if DEBUG_SWITCH_ld
#define print(str) std::cout << str << "\n";
#else
#define print(str) //
#endif

void printSet(std::unordered_set<controlNode *> set) {
    for (auto s : set) {
        print("node bb: " << s->BB->getNo());
        // std::cout<<"node bb: "<<s->BB->getNo()<<"\n";
    }
}

void LoopDetection::clear() {
    loops.clear();
    base2Loop.clear();
    loop2Base.clear();
    BB2Base.clear();
    stack.clear();
}

void LoopDetection::printloop(blockSet *loop) {
    for (auto bb : *loop) {
        print(" | bb:" << bb->getNo());
    }
}

BasicBlock *LoopDetection::getCondbb(BasicBlock *loopbb) {
    // if((loopbb->getPredBlock()).size()!=2)assert(0);
    for (auto pred : loopbb->getPredBlock()) {
        // print("== pred: "<<pred->getNo());
        if (pred != loopbb)
            return pred;
    }
}

BasicBlock *LoopDetection::getOutbb(BasicBlock *condbb) {
    auto head = condbb->end();
    CondBrInstruction *cond;
    for (auto instr = head->getNext(); instr != head;
         instr = instr->getNext()) {
        if (instr->isCond()) {
            cond = dynamic_cast<CondBrInstruction *>(instr);
            break;
        }
    }
    return cond->getFalseBranch();
}

BasicBlock *LoopDetection::getEndbb(BasicBlock *loopbb) {
    // if((loopbb->gketSuccBlock()).size()!=1)assert(0);
    // if((loopbb->getPredBlock()).size()!=2)assert(0);
    for (auto succ : loopbb->getSuccBlock()) {
        // print("== succ: "<<succ->getNo());
        if (succ != loopbb)
            return succ;
    }
}

void test(Unit *unit) {
    auto detector = new LoopDetection(unit);
    print("start run");

    detector->run();
    print("run finish");
    // areaCount = 0;
    for (auto it = detector->begin(); it != detector->end(); it++) {
        if (!detector->isBaseLoop(*it))
            continue;
        auto loop = *it;
        print("size = " << loop->size());
        detector->printloop(loop);
        BasicBlock *loopbb;
        BasicBlock *condbb;
        BasicBlock *endbb;
        if (loop->size() == 1) {
            BasicBlock *loopbb = *(loop->begin());
            condbb = detector->getCondbb(loopbb);
            endbb = detector->getEndbb(loopbb);
        } else {
            condbb = detector->getLoopCond(loop);
            endbb = detector->getOutbb(condbb);
        }
        print("-- condbb = " << condbb->getNo()
                             << " endbb = " << endbb->getNo());
    }
}

void LoopDetection::run() {

    for (auto func : unit->get_functions()) {
        for (auto bb : func->getBlockList()) {
            for (Instruction *ins = bb->begin(); ins != bb->end();
                 ins = ins->getNext()) {
                if (ins->isCond() &&
                    dynamic_cast<CondBrInstruction *>(ins)->getTrueBranch() ==
                        bb) {
                    // std::cout<<"??????? " << bb->getNo()<<"\n";
                    auto bset = new blockSet;
                    bset->insert(bb);
                    loops.insert(bset);
                }
            }
        }
    }
    // for (auto func : unit->get_functions()) {
    //     print("HANDLE: " << func->getFuncName());

    //     // clear();
    //     if (func->getBlockList().size() == 0)
    //         continue;

    //     buildCFG(func, nodes);
    //     print("buildCFG finish");

    //     while (stronglyConnComponents()) {
    //         if (comps.size() == 0)
    //             break;
    //         for (auto set : comps) {
    //             printSet(*set);
    //             auto base = findLoopBase(set, reserved);
    //             print("findLoopBase finish");
    //             // fix: 原因不明的空指针
    //             if(base==nullptr)return;
    //             // store result
    //             auto BBset = new blockSet;
    //             for (auto n : *set) {
    //                 BBset->insert(n->BB);
    //             }
    //             print("BBset->insert(n->BB);");
    //             loops.insert(BBset);
    //             print("loops insert finish" << base->BB->getNo());
    //             base2Loop.insert({base->BB, BBset});
    //             print("base2Loop insert finish");
    //             loop2Base.insert({BBset, base->BB});
    //             print("loop2Base insert finish");
    //             for (auto BB : *BBset) {
    //                 if (BB2Base.find(BB) == BB2Base.end())
    //                     BB2Base.insert({BB, base->BB});
    //                 else
    //                     BB2Base[BB] = base->BB;
    //             }
    //             // prepare for recongnize inner loop
    //             reserved.insert(base);
    //             nodes.erase(base);
    //             for (auto su : base->succs) {
    //                 su->preds.erase(base);
    //             }
    //             for (auto pred : base->preds) {
    //                 pred->succs.erase(base);
    //             }
    //         }

    //         for (auto set : comps){
    //             if(set)
    //             delete set;
    //         }
    //         comps.clear();
    //         for (auto n : nodes) {
    //             n->index = n->lowlink = -1;
    //             n->onStack = false;
    //         }
    //     }
    //     print("stronglyConnComponents finish");
    //     // clear
    //     for (auto n : reserved) {
    //         nodes.insert(n);
    //         for (auto pred : n->preds) {
    //             pred->succs.insert(n);
    //         }
    //     }
    //     reserved.clear();
    //     for (auto _n : nodes) {
    //         if(_n)
    //         delete _n;
    //     }
    //     nodes.clear();
    // }
}

bool LoopDetection::stronglyConnComponents() {
    idxcnt = 0;
    stack.clear();
    for (auto n : nodes) {
        if (n->index == -1)
            traverse(n);
    }
    return comps.size() != 0;
}

void LoopDetection::traverse(controlNode *n) {
    n->index = idxcnt++;
    n->lowlink = n->index;
    stack.push_back(n);
    n->onStack = true;

    for (auto su : n->succs) {
        if (su->index == -1) {
            traverse(su);
            n->lowlink = std::min(su->lowlink, n->lowlink);
        } else if (su->onStack) {
            n->lowlink = std::min(su->index, n->lowlink);
        }
    }

    if (n->index == n->lowlink) {
        auto set = new std::unordered_set<controlNode *>;
        controlNode *tmp;
        do {
            tmp = stack.back();
            tmp->onStack = false;
            set->insert(tmp);
            stack.pop_back();
        } while (tmp != n);
        if (set->size() == 1 && n->spin != true)
            delete set;
        else {
            // printSet(*set);
            comps.insert(set);
        }
    }
}

controlNode *
LoopDetection::findLoopBase(std::unordered_set<controlNode *> *set,
                            std::unordered_set<controlNode *> &reserved) {
    controlNode *base = nullptr;
    bool hadBeen = false;
    for (auto n : *set) {
        for (auto pred : n->preds) {
            if (set->find(pred) == set->end()) {
                base = n;
            }
        }
    }
    if (base != nullptr)
        return base;
    for (auto res : reserved) {
        for (auto succ : res->succs) {
            if (set->find(succ) != set->end()) {
                base = succ;
            }
        }
    }
    if (base == nullptr) {
        print("base==nullptr");
        // assert(0);
    }
    return base;
}

void LoopDetection::buildCFG(Function *func,
                             std::unordered_set<controlNode *> &result) {
    std::unordered_map<BasicBlock *, controlNode *> m;
    for (auto BB : func->getBlockList()) {
        auto node = new controlNode;
        node->BB = BB;
        node->index = node->lowlink = -1;
        node->onStack = false;
        m.insert({BB, node});
        result.insert(node);
    }
    for (auto BB : func->getBlockList()) {
        print("BB: " << BB->getNo()) auto node = m[BB];
        for (auto succ : BB->getSuccBlock()) {
            print("  BB succ: " << succ->getNo());
            node->succs.insert(m[succ]);
            if (succ == BB) {
                node->spin = true;
                // print("  Spin edge detected at node: " << BB->getNo())
            }
        }
        for (auto pred : BB->getPredBlock()) {
            print("  BB pred: " << pred->getNo());
            node->preds.insert(m[pred]);
        }
    }
}

std::string LoopDetection::printG(std::unordered_set<controlNode *> &nodes) {
    std::string nodeDecl;
    std::string eageDecl;
    std::vector<controlNode *> stack;
    std::unordered_set<controlNode *> processd;
    std::unordered_map<controlNode *, int> count;
    int counter = 0;
    for (auto n : nodes) {
        count.insert({n, counter++});
    }

    for (auto n : nodes) {
        for (auto succ : n->succs) {
            eageDecl += "\t" + std::to_string(count[n]) + "->" +
                        std::to_string(count[succ]) + "\n";
        }
    }
    int styleChoose;
    for (auto n : nodes) {
        styleChoose = 0;
        if (base2Loop.find(n->BB) != base2Loop.end()) {
            styleChoose = 1;
        }
    }
    return "digraph G {\n" + nodeDecl + "\n" + eageDecl + "}\n";
}

blockSet *LoopDetection::getBaseLoop(BasicBlock *BB) {
    if (BB2Base.find(BB) == BB2Base.end())
        return nullptr;
    return base2Loop[BB2Base[BB]];
}

BasicBlock *LoopDetection::getLoopCond(blockSet *set) {
    return getCondbb(loop2Base[set]);
}

blockSet *LoopDetection::getParentLoop(blockSet *set) {
    auto base = loop2Base[set];
    for (auto pred : base->getPredBlock()) {
        if (set->find(pred) != set->end())
            continue;
        auto loop = getBaseLoop(pred);
        if (loop == nullptr || loop->find(base) == loop->end())
            return nullptr;
        else {
            return loop;
        }
    }
    return nullptr;
}

bool LoopDetection::isBaseLoop(blockSet *set) {
    auto base = loop2Base[set];
    for (auto tmp : *set) {
        if (tmp == base)
            continue;
        if (base2Loop.find(tmp) != base2Loop.end())
            return false;
    }
    return true;
}

LoopDetection::~LoopDetection() {
    for (auto set : loops) {
        if (set)
            delete set;
    }
}

BasicBlock *LoopDetection::getLoopOut(blockSet *set) {
    std::unordered_set<BasicBlock *> OutBBs;
}
