#include "LoopStrengthReduction.h"

#define DEBUG_SWITCH_lsr 0
#if DEBUG_SWITCH_lsr
#define print(str) std::cout << str << "\n";
#else
#define print(str) //
#endif

#define NUM 25
#define NUMSECOND 1

bool isSrem = 0;
Instruction *bin_mod = nullptr;
Operand *cons = nullptr;
bool done = 0;
Instruction *findinst(int insttype, BasicBlock *bb) {
    for (Instruction *ins = bb->begin(); ins != bb->end();
         ins = ins->getNext()) {
        if (ins->getInstType() == insttype) {
            return ins;
        }
    }
    return nullptr;
}

void LoopStrengthReduction::execute() {
    print("======== LoopStrengthReduction START ========");
    detector = new LoopDetection(unit);
    detector->run();
    for (auto it = detector->begin(); it != detector->end(); it++) {
        // if (!detector->isBaseLoop(*it))
        //     continue;
        // if ((*it)->size() > 1) // 找自旋体
        //     continue;
        auto loop = *it;
        loopAccumulate2Mul(loop);
    }
    AccumInstEleVec.clear();
    defInLoop.clear();
    useInLoop.clear();
    LoopExpression.clear();
    OldInstrs.clear();
    if (done) {
        BlockMerge bm0(unit);
        bm0.execute();
        if (bm0.getChange()) {
            execute2();
        }
    }
    done = 0;

    print("======== LoopStrengthReduction END ========");
}
void LoopStrengthReduction::execute2() {
    print("======== LoopStrengthReduction START ========");
    detector = new LoopDetection(unit);
    detector->run();
    for (auto it = detector->begin(); it != detector->end(); it++) {
        // if (!detector->isBaseLoop(*it))
        //     continue;
        // if ((*it)->size() > 1) // 找自旋体
        //     continue;
        auto loop = *it;
        arithmetic2Mul(loop);
        // loopAccumulate2Mul(loop);
    }
    print("======== LoopStrengthReduction END ========");
}
// void LoopStrengthReduction::execute() {
//     print("======== LoopStrengthReduction START ========");
//     detector = new LoopDetection(unit);
//     BlockMerge bm0(unit);
//     int count = 1;
//     do {
//         std::cout<<"count = "<<count<<"\n";
//         detector->run();
//         for (auto it = detector->begin(); it != detector->end(); it++) {
//             // if (!detector->isBaseLoop(*it))
//             //     continue;
//             // if ((*it)->size() > 1) // 找自旋体
//             //     continue;
//             auto loop = *it;
//             loopAccumulate2Mul(loop);
//         }
//         bm0.execute();
//     } while (bm0.getChange())
//         print("======== LoopStrengthReduction END ========");
// }

bool LoopStrengthReduction::isDefInLoop(SymbolEntry *se) {
    if (std::find(defInLoop.begin(), defInLoop.end(), se) != defInLoop.end()) {
        print("   " << se->toStr() << " 1  ");
        return true;
    } else {
        print("   " << se->toStr() << " 0  ");
        return false;
    }
}

// 这应该是有点问题的
bool LoopStrengthReduction::isSelfBinConst(BinaryInstruction *inst) {
    Operand *var;
    Operand *constant;
    Operand *lval = inst->getDef();

    print("=== isSelfBinConst ===");
    print(inst->getInstName() << " " << inst->getUse()[0]->toStr() << " "
                              << inst->getUse()[1]->toStr());
    if (!inst->getUse()[0]->getSymbolEntry()->isConstant() &&
        inst->getUse()[1]->getSymbolEntry()->isConstant()) {
        var = inst->getUse()[0];
        constant = inst->getUse()[1];

    } else if (!inst->getUse()[1]->getSymbolEntry()->isConstant() &&
               inst->getUse()[0]->getSymbolEntry()->isConstant()) {
        var = inst->getUse()[1];
        constant = inst->getUse()[0];
    } else {
        return false;
    }
    if (var == constant)
        return false;
    auto dst = var->getDef();
    if (!dst->isPhi())
        return false;
    // AccumInst_phi = dst;
    print("== getVar def ==");
    bool sucess = false;
    PhiInstruction *phi = dynamic_cast<PhiInstruction *>(dst);
    for (auto src : phi->getSrcs()) {
        if (src.first == inst->getParent() && src.second == lval) {
            sucess = true;
            AccumEle = constant;
        }
    }
    return sucess;
}

bool LoopStrengthReduction::isSelfBin(BinaryInstruction *inst) {
    Operand *var;
    Operand *constant;
    Operand *lval = inst->getDef();
    auto opcode = inst->getOpcode();
    isSrem = 0;
    bin_mod = nullptr;
    SymbolEntry *var_se = nullptr;

    if (!(opcode == BinaryInstruction::SUB ||
          opcode == BinaryInstruction::ADD)) {
        stop = true;
        return false;
    }
    print("=== isSelfBin ===");
    print(inst->getInstName() << " " << inst->getUse()[0]->toStr() << " "
                              << inst->getUse()[1]->toStr());
    if (inst->getOpse().size() < 2)
        return 0;
    print("inst->getOpse() " << inst->getOpse()[0]->toStr() << " "
                             << inst->getOpse()[1]->toStr());
    if (!inst->getUse()[0]->getSymbolEntry()->isConstant() &&
        ((inst->getUse()[1]->getSymbolEntry()->isConstant()) ||
         !isDefInLoop(inst->getOpse()[1]) || inst->getUse()[1] == cons)) {
        var = inst->getUse()[0];
        var_se = inst->getOpse()[0];
        constant = inst->getUse()[1];

    } else if (!inst->getUse()[1]->getSymbolEntry()->isConstant() &&
               ((inst->getUse()[0]->getSymbolEntry()->isConstant()) ||
                !isDefInLoop(inst->getOpse()[0]) ||
                inst->getUse()[0] == cons)) {
        var_se = inst->getOpse()[1];
        var = inst->getUse()[1];
        constant = inst->getUse()[0];
    } else {
        return false;
    }
    if (var == constant)
        return false;
    auto dst = var->getDef();
    if (!dst)
        return false;
    print("var = " << var->toStr() << " const = " << constant->toStr());
    print(dst->getInstName() << " " << dst->getDef()->toStr());
    // if (!dst->isPhi())
    //     return false;

    // AccumInst_phi = dst;
    print("== getVar def ==");

    if (dst->isPhi()) {
        bool sucess = false;
        print("dst->isphi");
        PhiInstruction *phi = dynamic_cast<PhiInstruction *>(dst);
        for (auto src : phi->getSrcs()) {
            if (src.first == inst->getParent() && src.second == lval) {
                sucess = true;
                AccumInstEleVec.push_back(
                    std::make_pair(inst, std::make_pair(var, constant)));
                // AccumEle = constant;
            }
        }
        // isSrem out?
        // std::cout<<"isSrem? "<<inst->getOpse()[0]->toStr()<<"
        // "<<var_se->toStr()<<"\n";
        if (!sucess && inst->getDef()->getUse()[0]->isBin()) {
            BinaryInstruction *bin =
                dynamic_cast<BinaryInstruction *>(inst->getDef()->getUse()[0]);
            if (bin->getOpse()[0] != var_se)
                return false;
            if (bin->getOpcode() == BinaryInstruction::MOD) {
                for (auto src : phi->getSrcs()) {
                    if (src.first == inst->getParent() &&
                        src.second == bin->getDef()) {
                        sucess = true;
                        isSrem = 1;
                        bin_mod = bin;
                        AccumInstEleVec.push_back(std::make_pair(
                            inst, std::make_pair(var, constant)));
                    }
                }
            }
        }
        return sucess;
    } else if (dst->isLoad()) {
        // if(dst->getParent()!=var->getDef)

    } else {
        return false;
        // assert(0);
    }
}

int LoopStrengthReduction::getCmpInfo(unsigned opcode, double start,
                                      double step, double end,
                                      double &finalValue) {
    int times = 0;
    switch (opcode) {
    case CmpInstruction::E:
        print(start << " " << step << " E " << end);
        times = 1;
        finalValue = start + step;
        break;
    case CmpInstruction::NE:
        print(start << " " << step << " NE " << end);
        times = 0; // TODO：原因是不知道怎么算，目前打算直接条件判断掉
        finalValue = end;
        break;
    case CmpInstruction::L:
        print(start << " " << step << " L " << end);
        times = (end - start) / step;
        finalValue = end;
        break;
    case CmpInstruction::LE:
        print(start << " " << step << " LE " << end);
        times = (end - start) / step;
        if (step != 1 && (int)end % (int)step == 0)
            times += 1;
        finalValue = end + step;
        break;
    case CmpInstruction::G:
        print(start << " " << step << " G " << end);
        times = (start - end) / step;
        finalValue = end;
        break;
    case CmpInstruction::GE:
        print(start << " " << step << " GE " << end);
        times = (start - end) / step;
        if (step != 1 && (int)end % (int)step == 0)
            times += 1;
        finalValue = end - step;
        break;
    default:
        print("getCmpInfo sth wrong");
        assert(0);
        break;
    }
    return times;
}

void LoopStrengthReduction::getAllDefUse(BasicBlock *bb) {
    print("=== getAllDefUse ===");
    auto head = bb->end();
    SymbolEntry *se;
    for (auto instr = head->getNext(); instr != head;
         instr = instr->getNext()) {
        if (instr->isStore()) {
            se = dynamic_cast<IdentifierSymbolEntry *>(instr->getOpse()[0]);
            if (se == nullptr)
                continue;
            print(se->toStr());
            defInLoop.insert(se);
        } else if (instr->isPhi() && instr->getDst()->usersNum() > 0) {
            if (instr->getOpse()[0]->isTemporary())
                se = dynamic_cast<TemporarySymbolEntry *>(instr->getOpse()[0]);
            else
                se = dynamic_cast<IdentifierSymbolEntry *>(instr->getOpse()[0]);
            if (se == nullptr)
                continue;
            print("phi: " << se->toStr());
            defInLoop.insert(se);
        }
        // for (auto op : instr->getUse()) {
        //     if (op->isGlobal() &&
        //         useInLoop.find(op->getSymbolEntry()) == useInLoop.end()) {
        //         useInLoop.insert(op->getSymbolEntry());
        //     } else if (op->getOpdSe() &&
        //                useInLoop.find(op->getOpdSe()) == useInLoop.end()) {
        //         useInLoop.insert(op->getOpdSe());
        //     }
        // }
    }
}

BinaryInstruction *LoopStrengthReduction::binCheck(BinaryInstruction *bin) {
    // Instruction *res = nullptr;
    // for (auto use : bin->getDef()->getUse()) {
    //     // if (!(use->toStr()==0))
    //     //     res = use;
    // }
    // if (!res && res->isBin()) {
    //     return dynamic_cast<BinaryInstruction *>(res);
    // }
    // return nullptr;
}
void LoopStrengthReduction::mulAddDiv(Instruction *mul, BasicBlock *bb) {
    print("mulAddDiv");
    Operand *var1 = nullptr;
    Operand *constant1 = nullptr;
    for (auto op : mul->getUse()) {
        if (op->isConst())
            constant1 = op;
        else
            var1 = op;
    }
    if (constant1 == nullptr || var1 == nullptr)
        return;
    if (mul->getDef()->getUse().size() == 1 &&
        mul->getDef()->getUse()[0]->isBin()) {
        auto bin =
            dynamic_cast<BinaryInstruction *>(mul->getDef()->getUse()[0]);
        if (bin->getOpcode() == BinaryInstruction::ADD) {
            Operand *var = nullptr;
            Operand *constant = nullptr;
            for (auto op : bin->getUse()) {
                if (op->isConst())
                    constant = op;
                else
                    var = op;
            }
            // ADD指令一个常量一个变量
            if (constant &&
                dynamic_cast<ConstantSymbolEntry *>(constant->getSymbolEntry())
                        ->getValue() == 0 &&
                var != nullptr) {
                print(
                    "bin->getUse().size = " << bin->getDef()->getUse().size());
                // for(auto use:bin->getUse()){
                //     print(use->getDef()->toStr());
                // }
                print(bin->getDef()->toStr());
                print(bin->getDef()->getUse()[0]->getInstName());
                if (bin->getDef()->getUse().size() == 1 &&
                    bin->getDef()->getUse()[0]->isBin()) {
                    BinaryInstruction *bin2 = dynamic_cast<BinaryInstruction *>(
                        bin->getDef()->getUse()[0]);
                    // BinaryInstruction *bin2 = binCheck(bin);
                    print(bin2->getDef()->toStr());
                    if (bin2->getOpcode() == BinaryInstruction::DIV) {
                        Operand *var2 = nullptr;
                        Operand *constant2 = nullptr;
                        for (auto op : mul->getUse()) {
                            if (op->isConst())
                                constant2 = op;
                            else
                                var2 = op;
                        }
                        if (dynamic_cast<ConstantSymbolEntry *>(
                                constant1->getSymbolEntry())
                                ->getValue() ==
                            dynamic_cast<ConstantSymbolEntry *>(
                                constant2->getSymbolEntry())
                                ->getValue()) {
                            bin2->getDef()->replaceAllUseWith(var1);
                            bb->remove(mul);
                            bb->remove(bin);
                            bb->remove(bin2);
                        }
                    }
                }
            }
        }
    }
}

void LoopStrengthReduction::eliUselessInsts(BasicBlock *bb) {
    print("============== start eliUselessInsts ================");
    std::set<Instruction *> removeList = {};
    for (auto instr = bb->begin(); instr != bb->end();
         instr = instr->getNext()) {
        if (instr->getDef()->getUse().size() == 0) {
            removeList.insert(instr);
            continue;
        }
        if (instr->isPhi()) {
            auto phi = dynamic_cast<PhiInstruction *>(instr);
            if (phi->getSrcs().size() == 1) {
                phi->getDef()->replaceAllUseWith(
                    (*phi->getSrcs().begin()).second);
                removeList.insert(instr);
                continue;
            }
        }
    }
    for (auto inst : removeList) {
        for (auto use : inst->getUse()) {
            use->removeUse(inst);
        }
        bb->remove(inst);
    }
    for (auto instr = bb->begin(); instr != bb->end();
         instr = instr->getNext()) {
        if (instr->getDef()->getUse().size() == 0) {
            removeList.insert(instr);
            continue;
        }
        if (instr->isBin() &&
            dynamic_cast<BinaryInstruction *>(instr)->getOpcode() ==
                BinaryInstruction::MUL) {
            mulAddDiv(instr, bb);
        }
    }
}

bool isArithmetic(IdentifierSymbolEntry *funcSe) {
    if (funcSe->isSysy() || funcSe->getName() == "llvm.memset.p0.i32")
        return 0;
    else
        return 1;
}

// 只处理d = i
void LoopStrengthReduction::arithmetic2Mul(blockSet *loop) {
    cons = nullptr;
    print("============== start arithmetic2Mul ================");
    for (auto bb : *loop) {
        eliUselessInsts(bb);
        print("end eliUselessInsts ") CmpInstruction *cmp =
            dynamic_cast<CmpInstruction *>(findinst(Instruction::CMP, bb));
        if (cmp->getOpcode() == CmpInstruction::NE)
            return;
        CallInstruction *call =
            dynamic_cast<CallInstruction *>(findinst(Instruction::CALL, bb));
        if (call) {
            IdentifierSymbolEntry *funcSe =
                dynamic_cast<IdentifierSymbolEntry *>(call->getSymEntry());
            if (!isArithmetic(funcSe))
                return;
            cons = call->getDef();
        } else {
            return;
        }
        print("call");
        auto se1_op = cmp->getOp()[1];
        auto se1 = se1_op->getSymbolEntry();
        auto se2_op = cmp->getOp()[2];
        auto se2 = se2_op->getSymbolEntry();
        print("se1: " << se1->toStr() << " se2: " << se2->toStr());

        if (!se2->isConstant() && !se1->isConstant()) {
            print("== se2->isVariable() && se1->isVariable() ==");
            if (!se1->getType()->isInt())
                break;
            loopController = se1_op; // Todo: 翻转后注意变一下
            loopController_se = se1; // Todo: 翻转后注意变一下
            auto bin = se1_op->getDef();

            if (bin->isBin()) { // Todo:如果是全局变量就该处理load
                // print(se1_op->toStr());
                auto src1 = bin->getUse()[0];
                auto src2 = bin->getUse()[1];
                // todo:条件先这样写死一点
                print("srcs: " << src1->toStr() << " " << src2->toStr());
                print(src1->getSymbolEntry()->toStr()
                      << " " << src2->getSymbolEntry()->toStr());
                if (!src1->getSymbolEntry()->isConstant() &&
                    src2->getSymbolEntry()->isConstant()) {
                    print("==== src1 var & src2 const ====");
                    // 意思就是只想做单循环控制变量的pass
                    if (!src1->getDef()->isPhi())
                        return;
                    print("src1->getUse().size() = " << src1->getUse().size());
                    // 不接受控制变量控制其他变量
                    // if (src1->getUse().size() > 2)
                    //     return;
                    auto phi = dynamic_cast<PhiInstruction *>(src1->getDef());
                    loopControlInst_phi = phi;

                    print("======== findExpression ========");
                    findExpression(bb);

                    // if (AccumInst == nullptr) {
                    //     print("no AccumInst");
                    //     break;
                    // }

                    if (AccumInstEleVec.size() == 0) {
                        print("no Accum.");
                        break;
                    }

                    print("== get AccumInst ==");
                    // 找控制变量初值
                    double start = 0;
                    for (auto src : phi->getSrcs()) {
                        print("B" << src.first->getNo() << " "
                                  << src.second->toStr());

                        if (src.first != bb && src.second->isConst()) {
                            start = dynamic_cast<ConstantSymbolEntry *>(
                                        src.second->getSymbolEntry())
                                        ->getValue();
                            break;
                        }
                    }
                    print("== get controller initVal ==");
                    double step = dynamic_cast<ConstantSymbolEntry *>(
                                      src2->getSymbolEntry())
                                      ->getValue();
                    // Todo: 先这样了
                    if (step != 1 || start != 0)
                        break;
                    Operand *end = se2_op;
                    updateBlock_ari(cmp->getOpcode(), bb, start, step, end);
                    // double finalValue = 0;

                    // int times = getCmpInfo(cmp->getOpcode(), start, step,
                    //                        end, finalValue);
                    // print("== updateBlock ==");
                    // updateBlock(bb, times, finalValue);
                }
            }
        }
    }
}

void LoopStrengthReduction::loopAccumulate2Mul(blockSet *loop) {
    print("============== start loopAccumulate2Mul ================");
    std::set<Operand *> useInLoop = {};
    for (auto bb : *loop) {
        loopController = nullptr;
        // Todo: 先阉割下数组
        // if (findinst(Instruction::GEP, bb) != nullptr)
        //     break;
        bool exit = false;
        auto head = bb->end();
        for (auto instr = head->getNext(); instr != head;
             instr = instr->getNext()) {
            if (instr->isGep() ||
                instr->isCall() &&
                    dynamic_cast<IdentifierSymbolEntry *>(
                        dynamic_cast<CallInstruction *>(instr)->getFunc_se())
                        ->isSysy()) {
                exit = true;
            }
        }
        if (exit)
            break;
        if (findinst(Instruction::CMP, bb) != nullptr) {
            print("======== find cmp ========");
            // auto head = bb->end();
            // for (auto instr = head->getNext(); instr != head;
            //      instr = instr->getNext()) {
            //     // fix：按照cmp在末尾考虑的，有可能需要改
            //     if (instr->isRet() || instr->isCall()) {
            //         break;
            //     }
            // if (instr&&instr->isCmp()) {
            // CmpInstruction *cmp = dynamic_cast<CmpInstruction *>(instr);
            CmpInstruction *cmp =
                dynamic_cast<CmpInstruction *>(findinst(Instruction::CMP, bb));
            // Todo: 先把不好做的卡掉吧
            if (cmp->getOpcode() == CmpInstruction::NE)
                break;
            auto se1_op = cmp->getOp()[1];
            auto se1 = se1_op->getSymbolEntry();
            auto se2_op = cmp->getOp()[2];
            auto se2 = se2_op->getSymbolEntry();
            print("se1: " << se1->toStr() << " se2: " << se2->toStr());
            /* se1变量 se2常量 */
            if (se2->isConstant() && !se1->isConstant()) {
                print("== se2->isConstant() && se1->isVariable() ==");
                if (!se1->getType()->isInt())
                    break;
                loopController = se1_op; // Todo: 翻转后注意变一下
                loopController_se = se1; // Todo: 翻转后注意变一下
                auto bin = se1_op->getDef();

                if (bin->isBin()) { // Todo:如果是全局变量就该处理load
                    // print(se1_op->toStr());
                    auto src1 = bin->getUse()[0];
                    auto src2 = bin->getUse()[1];
                    // todo:条件先这样写死一点
                    print("srcs: " << src1->toStr() << " " << src2->toStr());
                    print(src1->getSymbolEntry()->toStr()
                          << " " << src2->getSymbolEntry()->toStr());
                    if (!src1->getSymbolEntry()->isConstant() &&
                        src2->getSymbolEntry()->isConstant()) {
                        print("==== src1 var & src2 const ====");
                        // 意思就是只想做单循环控制变量的pass
                        if (!src1->getDef()->isPhi())
                            break;
                        // 不接受控制变量控制其他变量
                        if (src1->getUse().size() > 2)
                            break;
                        auto phi =
                            dynamic_cast<PhiInstruction *>(src1->getDef());
                        loopControlInst_phi = phi;

                        print("======== findExpression ========");
                        findExpression(bb);

                        // if (AccumInst == nullptr) {
                        //     print("no AccumInst");
                        //     break;
                        // }

                        if (AccumInstEleVec.size() == 0) {
                            print("no Accum.");
                            break;
                        }

                        print("== get AccumInst ==");
                        double start = 0;
                        for (auto src : phi->getSrcs()) {
                            print("B" << src.first->getNo() << " "
                                      << src.second->toStr());
                            if (src.first != bb && src.second->isConst()) {
                                start = dynamic_cast<ConstantSymbolEntry *>(
                                            src.second->getSymbolEntry())
                                            ->getValue();
                                break;
                            }
                        }
                        print("== get controller initVal ==");
                        double step = dynamic_cast<ConstantSymbolEntry *>(
                                          src2->getSymbolEntry())
                                          ->getValue();
                        double end = dynamic_cast<ConstantSymbolEntry *>(se2)
                                         ->getValue();
                        double finalValue = 0;

                        int times = getCmpInfo(cmp->getOpcode(), start, step,
                                               end, finalValue);
                        print("== updateBlock ==");
                        updateBlock(bb, times, finalValue);
                    }
                }
            } else if (
                // se2->isConstant() && // levelup
                !se1->isConstant()) {
                print("== se2->isVariable() && se1->isVariable() ==");
                if (!se1->getType()->isInt())
                    break;
                loopController = se1_op; // Todo: 翻转后注意变一下
                loopController_se = se1; // Todo: 翻转后注意变一下
                auto bin = se1_op->getDef();

                if (bin->isBin()) { // Todo:如果是全局变量就该处理load
                    // print(se1_op->toStr());
                    auto src1 = bin->getUse()[0];
                    auto src2 = bin->getUse()[1];
                    // todo:条件先这样写死一点
                    print("srcs: " << src1->toStr() << " " << src2->toStr());
                    print(src1->getSymbolEntry()->toStr()
                          << " " << src2->getSymbolEntry()->toStr());
                    if (!src1->getSymbolEntry()->isConstant() &&
                        src2->getSymbolEntry()->isConstant()) {
                        print("==== src1 var & src2 const ====");
                        // 意思就是只想做单循环控制变量的pass
                        if (!src1->getDef()->isPhi())
                            break;
                        // 不接受控制变量控制其他变量
                        if (src1->getUse().size() > 2)
                            break;
                        auto phi =
                            dynamic_cast<PhiInstruction *>(src1->getDef());
                        loopControlInst_phi = phi;

                        print("======== findExpression ========");
                        findExpression(bb);

                        // if (AccumInst == nullptr) {
                        //     print("no AccumInst");
                        //     break;
                        // }

                        if (AccumInstEleVec.size() == 0) {
                            print("no Accum.");
                            break;
                        }

                        print("== get AccumInst ==");
                        // 找控制变量初值
                        double start = 0;
                        for (auto src : phi->getSrcs()) {
                            print("B" << src.first->getNo() << " "
                                      << src.second->toStr());

                            if (src.first != bb && src.second->isConst()) {
                                start = dynamic_cast<ConstantSymbolEntry *>(
                                            src.second->getSymbolEntry())
                                            ->getValue();
                                break;
                            }
                        }
                        print("== get controller initVal ==");
                        double step = dynamic_cast<ConstantSymbolEntry *>(
                                          src2->getSymbolEntry())
                                          ->getValue();
                        // Todo: 先这样了
                        if (step != 1 || start != 0)
                            break;
                        Operand *end = se2_op;
                        updateBlock_var(cmp->getOpcode(), bb, start, step, end);
                        // double finalValue = 0;

                        // int times = getCmpInfo(cmp->getOpcode(), start, step,
                        //                        end, finalValue);
                        // print("== updateBlock ==");
                        // updateBlock(bb, times, finalValue);
                    }
                }
            }
            /* se2变量 se1常量 */
            // if(se2->isConstant()&&!se1->isConstant())
            // else if(instr->isCond()){ // 默认在末尾
        }
        // else {
        //     for (auto op : instr->getUse()) {
        //         if (useInLoop.find(op) == useInLoop.end()) {
        //             useInLoop.insert(op);
        //         }
        //     }
        // }
        // }
    }
    print("============== end loopAccumulate2Mul ================");
}

// void LoopStrengthReduction::markUse(Instruction *instr) {
//     // if (instr->getDef())
//     //     print("  | " << instr->getInstName() << " "
//     //                  << instr->getOp()[0]->toStr());
//     // print(&instr);

//     if (LoopExpression.find(instr) != LoopExpression.end())
//         return;
//     LoopExpression.insert(instr);

//     if (instr->isStore() || instr->isPhi()) {
//         print("这里不该有phi/store");
//         assert(0);
//     } else {
//         for (auto op : instr->getUse()) {
//             // if (worklist_op.find(op->toStr()) == worklist_op.end())
//             //     worklist_op.insert(op->toStr());
//             if (op->getDef()) {
//                 // print(op->getDef()->getInstName());
//                 markUse(op->getDef());
//             }
//         }
//     }
// }

/** Todo: 看看有没有可能做一下这种:
 * x = x + i1 + i2 + i3;
 * x = x + i1 + i2 + i3;
 * finish.
 */
void LoopStrengthReduction::findExpression(BasicBlock *bb) {
    getAllDefUse(bb);
    stop = false;

    // AccumInst_phi = nullptr;
    // AccumInst = nullptr;
    LoopExpression.clear();

    auto head = bb->end();
    for (auto instr = head->getNext(); instr != head;
         instr = instr->getNext()) {
        OldInstrs.insert(instr);
        if (instr->isBin() && instr->getDef() != loopController) {
            LoopExpression.insert(instr);
        }
    }
    // for (auto instr : LoopExpression) {
    // markUse(instr)
    // }
    for (auto instr : LoopExpression) {
        // if (isSelfBinConst((BinaryInstruction *)instr)) {
        if (isSelfBin((BinaryInstruction *)instr)) {
            // AccumInst = instr;
            break;
        }
    }

    if (stop)
        AccumInstEleVec.clear();
    // Todo: 看看有没有人拿Accum当右值
    // for (auto pair : AccumInstEleVec) {

    // }
}

// 循环次数未知，非常量
void LoopStrengthReduction::updateBlock_var(int opcode, BasicBlock *bb,
                                            int start, int step, Operand *end) {
    print("======== updateBlock_var ========");
    Operand *start_op =
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, start));
    Operand *step_op =
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, step));
    Operand *times_op = nullptr;
    Operand *finalValue = nullptr;
    if (start == 0 && step == 1) {
        times_op = end;
        finalValue = end;
    } else {
        switch (opcode) {
            Operand *dst;
        case CmpInstruction::L:
            // times = (end - start) / step;

            dst = new Operand(new TemporarySymbolEntry(
                TypeSystem::intType, SymbolTable::getLabel()));
            new BinaryInstruction(BinaryInstruction::SUB, dst, end, start_op,
                                  bb);

            times_op = new Operand(new TemporarySymbolEntry(
                TypeSystem::intType, SymbolTable::getLabel()));
            new BinaryInstruction(BinaryInstruction::SUB, times_op, dst,
                                  step_op, bb);

            finalValue = end;
            break;
        case CmpInstruction::G:
            // times = (start - end) / step;
            dst = new Operand(new TemporarySymbolEntry(
                TypeSystem::intType, SymbolTable::getLabel()));
            new BinaryInstruction(BinaryInstruction::SUB, dst, start_op, end,
                                  bb);

            times_op = new Operand(new TemporarySymbolEntry(
                TypeSystem::intType, SymbolTable::getLabel()));
            new BinaryInstruction(BinaryInstruction::SUB, times_op, dst,
                                  step_op, bb);

            finalValue = end;
            break;
        default:
            print("invelid opcode");
            return;
        }
    }

    auto cond =
        dynamic_cast<CondBrInstruction *>(findinst(Instruction::COND, bb));
    if (end->getDef()->isLoad()) {
        new LoadInstruction(end->getDef()->getDef(), end->getDef()->getUse()[0],
                            bb);
    }

    Operand *terminal = nullptr;
    for (auto pair : AccumInstEleVec) {
        /* 处理含取模的累加 */
        if (isSrem != 0) {
            // 拆开多块取模
            auto num =
                new Operand(new ConstantSymbolEntry(TypeSystem::intType, NUM));
            auto dst_div = new Operand(new TemporarySymbolEntry(
                TypeSystem::intType, SymbolTable::getLabel()));
            new BinaryInstruction(BinaryInstruction::DIV, dst_div, times_op,
                                  num, bb);

            Operand *opd;
            auto def = (pair.second).first->getDef();
            if (def->isPhi()) {
                for (auto src :
                     dynamic_cast<PhiInstruction *>(def)->getSrcs()) {
                    if (src.first != bb) {
                        opd = src.second;
                    }
                }
            } else if (def->isLoad()) {
                opd = (pair.second).first;
            }

            // 切割NUM份
            auto addr = new Operand(new TemporarySymbolEntry(
                TypeSystem::intType, SymbolTable::getLabel()));
            new BinaryInstruction(BinaryInstruction::MUL, addr,
                                  (pair.second).second, dst_div, bb);
            Operand *dst;
            for (int i = 0; i < NUM; i++) {
                // 最后一个def要接上
                if (i == NUM - 1) {
                    dst = bin_mod->getDef();
                } else {
                    dst = new Operand(new TemporarySymbolEntry(
                        TypeSystem::intType, SymbolTable::getLabel()));
                }

                auto dst_srem = new Operand(new TemporarySymbolEntry(
                    TypeSystem::intType, SymbolTable::getLabel()));

                // 累加变量
                new BinaryInstruction(BinaryInstruction::ADD, dst_srem, opd,
                                      addr, bb);

                new BinaryInstruction(BinaryInstruction::MOD, dst, dst_srem,
                                      bin_mod->getUse()[1], bb);

                print("====================== " << dst->toStr());

                opd = dst;
            }

        } else { /* 普通累加  */
            // 累加变量
            auto addr = new Operand(new TemporarySymbolEntry(
                TypeSystem::intType, SymbolTable::getLabel()));
            new BinaryInstruction(BinaryInstruction::MUL, addr,
                                  (pair.second).second, times_op, bb);

            auto def = (pair.second).first->getDef();
            Operand *opd;
            if (def->isPhi()) {
                for (auto src :
                     dynamic_cast<PhiInstruction *>(def)->getSrcs()) {
                    if (src.first != bb) {
                        opd = src.second;
                    }
                }
            } else if (def->isLoad()) {
                opd = (pair.second).first;
            }
            // pair.first->getDef()
            //     ->getUse_v()
            //     .clear(); // 原来的东西咋没自动删掉啊
            // print("pair.first->getDef()-> "<<pair.first->getDef()->toStr());
            new BinaryInstruction(pair.first->getOpcode(), pair.first->getDef(),
                                  opd, addr, bb);
        }
        terminal = pair.first->getDef();
    }

    // // 额外优化循环内的取模变量
    // if (isSrem != 0) {
    //     // 溢出了，总之先MOD
    //     // 但是后续可能产生问题，不确定MOD次数对数字大小来说是否足够

    //     new BinaryInstruction(BinaryInstruction::MOD, bin_mod->getDef(),
    //                           terminal, bin_mod->getUse()[1], bb);
    // }

    auto zero = new Operand(new ConstantSymbolEntry(TypeSystem::intType, 0));

    // 控制变量
    BinaryInstruction *loopControllerFinalVal = new BinaryInstruction(
        BinaryInstruction::ADD, loopController, finalValue, zero, bb);
    // cond -> uncond
    new UncondBrInstruction(
        dynamic_cast<CondBrInstruction *>(cond)->getFalseBranch(), bb);

    for (auto inst : OldInstrs) {
        bb->remove(inst);
        delete inst;
    }
    OldInstrs.clear();

    // 维护前驱后继
    auto succs = bb->getSuccBlock();
    if (std::find(succs.begin(), succs.end(), bb) != succs.end())
        bb->removeSucc(bb);

    auto preds = bb->getPredBlock();
    if (std::find(preds.begin(), preds.end(), bb) != preds.end())
        bb->removePred(bb);
    done = 1;
    std::cout << "yes lsr done.\n";
}

void LoopStrengthReduction::updateBlock(BasicBlock *bb, int times,
                                        int finalValue) {
    print("======== updateBlock ========");
    // assert(AccumInst != nullptr);
    auto cond =
        dynamic_cast<CondBrInstruction *>(findinst(Instruction::COND, bb));
    auto times_op =
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, times));

    print("普通普通普通普通普通普通普通");
    for (auto pair : AccumInstEleVec) {
        // 累加变量
        auto addr = new Operand(new TemporarySymbolEntry(
            TypeSystem::intType, SymbolTable::getLabel()));
        new BinaryInstruction(BinaryInstruction::MUL, addr,
                              (pair.second).second, times_op, bb);
        auto def = (pair.second).first->getDef();
        Operand *opd;
        if (def->isPhi()) {
            for (auto src : dynamic_cast<PhiInstruction *>(def)->getSrcs()) {
                if (src.first != bb) {
                    opd = src.second;
                }
            }
        } else if (def->isLoad()) {
            opd = (pair.second).first;
        }
        // pair.first->getDef()->getUse_v().clear();// 原来的东西咋没自动删掉啊
        print(pair.first->getDef()->toStr()
              << " " << pair.first->getDef()->getUse_v().size());
        new BinaryInstruction(pair.first->getOpcode(), pair.first->getDef(),
                              opd, addr, bb);
    }

    auto zero = new Operand(new ConstantSymbolEntry(TypeSystem::intType, 0));
    auto finalValue_op =
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, finalValue));

    // 控制变量
    BinaryInstruction *loopControllerFinalVal = new BinaryInstruction(
        BinaryInstruction::ADD, loopController, finalValue_op, zero, bb);
    // cond -> uncond
    new UncondBrInstruction(
        dynamic_cast<CondBrInstruction *>(cond)->getFalseBranch(), bb);

    for (auto inst : OldInstrs) {
        bb->remove(inst);
        delete inst;
    }
    OldInstrs.clear();

    // 维护前驱后继
    auto succs = bb->getSuccBlock();
    if (std::find(succs.begin(), succs.end(), bb) != succs.end())
        bb->removeSucc(bb);

    auto preds = bb->getPredBlock();
    if (std::find(preds.begin(), preds.end(), bb) != preds.end())
        bb->removePred(bb);

    done = 1;
    std::cout << "yes lsr done.\n";
}

void LoopStrengthReduction::updateBlock_ari(int opcode, BasicBlock *bb,
                                            int start, int step, Operand
                                            *end) {
    print("======== updateBlock_ari ========");
    Operand *start_op =
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, start));
    Operand *step_op =
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, step));
    Operand *times_op = nullptr;
    Operand *finalValue = nullptr;
    if (start == 0 && step == 1) {
        times_op = end;
        finalValue = end;
    } else {
        switch (opcode) {
            Operand *dst;
        case CmpInstruction::L:
            // times = (end - start) / step;

            dst = new Operand(new TemporarySymbolEntry(
                TypeSystem::intType, SymbolTable::getLabel()));
            new BinaryInstruction(BinaryInstruction::SUB, dst, end, start_op,
                                  bb);

            times_op = new Operand(new TemporarySymbolEntry(
                TypeSystem::intType, SymbolTable::getLabel()));
            new BinaryInstruction(BinaryInstruction::SUB, times_op, dst,
                                  step_op, bb);

            finalValue = end;
            break;
        case CmpInstruction::G:
            // times = (start - end) / step;
            dst = new Operand(new TemporarySymbolEntry(
                TypeSystem::intType, SymbolTable::getLabel()));
            new BinaryInstruction(BinaryInstruction::SUB, dst, start_op, end,
                                  bb);

            times_op = new Operand(new TemporarySymbolEntry(
                TypeSystem::intType, SymbolTable::getLabel()));
            new BinaryInstruction(BinaryInstruction::SUB, times_op, dst,
                                  step_op, bb);

            finalValue = end;
            break;
        default:
            print("invelid opcode");
            return;
        }
    }

    auto cond =
        dynamic_cast<CondBrInstruction *>(findinst(Instruction::COND, bb));
    if (end->getDef()->isLoad()) {
        new LoadInstruction(end->getDef()->getDef(),
        end->getDef()->getUse()[0],
                            bb);
    }

    auto condbb = bb->getPredBlock()[0];
    auto cond_prev =
        dynamic_cast<CondBrInstruction *>(findinst(Instruction::COND,
        condbb));
    print("cond done.");
    Operand *terminal = nullptr;
    auto two = new Operand(new ConstantSymbolEntry(TypeSystem::intType, 2));
    auto one = new Operand(new ConstantSymbolEntry(TypeSystem::intType, 1));
    auto zero = new Operand(new ConstantSymbolEntry(TypeSystem::intType, 0));

    BasicBlock *trueBranch = bb;
    BasicBlock *falseBranch = new BasicBlock(bb->getParent());
    auto cmpbb = new BasicBlock(bb->getParent());
    Operand *ld;
    print("end = " << end->toStr());
    if (end->getDef()->isLoad()) {
        ld = new Operand(new TemporarySymbolEntry(TypeSystem::boolType,
                                                  SymbolTable::getLabel()));
        new LoadInstruction(ld, end->getDef()->getUse()[0], cmpbb);
    } else {
        ld = end;
    }
    auto mod2dst = new Operand(
        new TemporarySymbolEntry(TypeSystem::intType,
        SymbolTable::getLabel()));
    new BinaryInstruction(BinaryInstruction::MOD, mod2dst, ld, two, cmpbb);
    auto cmpDst = new Operand(new TemporarySymbolEntry(
        TypeSystem::boolType, SymbolTable::getLabel()));
    new CmpInstruction(CmpInstruction::E, cmpDst, mod2dst, zero, cmpbb);
    new CondBrInstruction(trueBranch, falseBranch, cmpDst, cmpbb);

    for (auto b : bb->getPredBlock()) {
        cmpbb->addPred(b);
        b->addSucc(cmpbb);
    }
    print("cmpbb done.");
    bb->getPredBlock().clear();
    cmpbb->addSucc(bb);
    cmpbb->addSucc(falseBranch);
    bb->addPred(cmpbb);
    falseBranch->addPred(cmpbb);
    cond_prev->setTrueBranch(cmpbb);

    print("cmpbb relation done.");
    for (auto pair : AccumInstEleVec) {
        // 累加变量
        auto addr = new Operand(new TemporarySymbolEntry(
            TypeSystem::intType, SymbolTable::getLabel()));
        new BinaryInstruction(BinaryInstruction::DIV, addr, times_op, two,
        bb);

        auto addr2 = new Operand(new TemporarySymbolEntry(
            TypeSystem::intType, SymbolTable::getLabel()));
        new BinaryInstruction(BinaryInstruction::SUB, addr2, times_op, one,
        bb);

        auto addr3 = new Operand(new TemporarySymbolEntry(
            TypeSystem::intType, SymbolTable::getLabel()));
        new BinaryInstruction(BinaryInstruction::MUL, addr3, addr, addr2,
        bb);

        auto def = (pair.second).first->getDef();
        print((pair.second).first->toStr());
        Operand *opd;
        if (def->isPhi()) {
            for (auto src : dynamic_cast<PhiInstruction *>(def)->getSrcs()) {
                if (src.first != bb) {
                    opd = src.second;
                }
            }
        } else if (def->isLoad()) {
            opd = (pair.second).first;
        }
        // pair.first->getDef()->getUse_v().clear();
        new BinaryInstruction(pair.first->getOpcode(), pair.first->getDef(),
                              opd, addr3, bb);
        terminal = pair.first->getDef();
    }
    // if (end->getDef()->isLoad()) {
    //     new LoadInstruction(end->getDef()->getDef(),
    //     end->getDef()->getUse()[0],
    //                         falseBranch);
    // }
    print("truebb relation done.");
    Operand *ld2;
    if (end->getDef()->isLoad()) {
        ld2 = new Operand(new TemporarySymbolEntry(TypeSystem::boolType,
                                                   SymbolTable::getLabel()));
        new LoadInstruction(ld2, end->getDef()->getUse()[0], falseBranch);
    } else {
        ld2 = end;
    }
    Operand *terminal2 = new Operand(
        new TemporarySymbolEntry(TypeSystem::intType,
        SymbolTable::getLabel()));
    for (auto pair : AccumInstEleVec) {
        // 累加变量
        auto addr2 = new Operand(new TemporarySymbolEntry(
            TypeSystem::intType, SymbolTable::getLabel()));
        new BinaryInstruction(BinaryInstruction::SUB, addr2, ld2, one,
                              falseBranch);

        auto addr = new Operand(new TemporarySymbolEntry(
            TypeSystem::intType, SymbolTable::getLabel()));
        new BinaryInstruction(BinaryInstruction::DIV, addr, addr2, two,
                              falseBranch);

        auto addr3 = new Operand(new TemporarySymbolEntry(
            TypeSystem::intType, SymbolTable::getLabel()));
        new BinaryInstruction(BinaryInstruction::MUL, addr3, addr, ld2,
                              falseBranch);

        auto def = (pair.second).first->getDef();
        print((pair.second).first->toStr());
        Operand *opd;
        if (def->isPhi()) {
            for (auto src : dynamic_cast<PhiInstruction *>(def)->getSrcs()) {
                if (src.first != bb) {
                    opd = src.second;
                }
            }
        } else if (def->isLoad()) {
            opd = (pair.second).first;
        }
        // pair.first->getDef()->getUse_v().clear();
        new BinaryInstruction(pair.first->getOpcode(), terminal2, opd, addr3,
                              falseBranch);
        // terminal = pair.first->getDef();
    }
    print("AccumInstEleVec done.");

    new BinaryInstruction(BinaryInstruction::MOD, bin_mod->getDef(),
    terminal,
                          bin_mod->getUse()[1], bb);
    // auto sum = new Operand(
    //     new TemporarySymbolEntry(TypeSystem::intType,
    //     SymbolTable::getLabel()));
    new BinaryInstruction(BinaryInstruction::MOD,
                          bin_mod->getDef()->getUse()[1]->getDef(),
                          terminal2, bin_mod->getUse()[1], falseBranch);

    // // 控制变量
    // BinaryInstruction *loopControllerFinalVal = new BinaryInstruction(
    //     BinaryInstruction::ADD, loopController, finalValue, zero, bb);
    // cond -> uncond
    BasicBlock *endbb =
        dynamic_cast<CondBrInstruction *>(cond)->getFalseBranch();
    // auto tmp2 = new BasicBlock(bb->getParent());
    new UncondBrInstruction(endbb, bb);

    new UncondBrInstruction(endbb, falseBranch);

    // auto tmp = new BasicBlock(bb->getParent());
    // new UncondBrInstruction(tmp, falseBranch);
    // new UncondBrInstruction(endbb, tmp);
    // falseBranch->addSucc(tmp);
    // tmp->addPred(falseBranch);
    // tmp->addSucc(endbb);
    // endbb->addPred(tmp);

    // new UncondBrInstruction(endbb, tmp2);
    // bb->getSuccBlock().clear();
    // bb->addSucc(tmp2);
    // tmp2->addPred(bb);
    // tmp2->addSucc(endbb);
    // endbb->addPred(tmp2);

    // std::cout << "bin_mod->getDef() = " << bin_mod->getDef()->toStr() <<
    // ""
    //           << terminal->getUse().size() << "\n";
    // print("bin_mod->getDef() = " << bin_mod->getDef()->toStr() << " "
                                //  << terminal->getUse().size());
    // if (bin_mod->getDef()->getUse()[1]->isPhi()) {
    //     auto phi_t =
    //         dynamic_cast<PhiInstruction*>(bin_mod->getDef()->getUse()[1]);
    //     // std::cout << "8888 " << phi_t->getDef()->toStr() << "\n";
    //     // phi_t->addSrc(tmp, sum);
    //     // phi_t->addSrc(falseBranch, sum);
    //     // phi_t->addSrc(tmp2, terminal);
    // }

    for (auto inst : OldInstrs) {
        bb->remove(inst);
        delete inst;
    }
    OldInstrs.clear();

    // 维护前驱后继
    auto succs = bb->getSuccBlock();
    if (std::find(succs.begin(), succs.end(), bb) != succs.end())
        bb->removeSucc(bb);

    auto preds = bb->getPredBlock();
    if (std::find(preds.begin(), preds.end(), bb) != preds.end())
        bb->removePred(bb);
    done = 1;
    std::cout << "yes lsr-ari done.\n";
}

// void getLoopJudge(BasicBlock *bb, BasicBlock *succ, BasicBlock *pred,
//                   Operand *times_op) {
//     auto limit = new Operand(
//         new ConstantSymbolEntry(TypeSystem::intType, OVERFLOWLIMIT));
//     auto dst = new Operand(
//         new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
//     new BinaryInstruction(BinaryInstruction::MOD, mod2dst, ld, two, cmpbb);
//     BasicBlock *endbb =
//         dynamic_cast<CondBrInstruction *>(cond)->getFalseBranch();
//     auto cond =
//         dynamic_cast<CondBrInstruction *>(findinst(Instruction::COND, bb));
//     new LoadInstruction(end->getDef()->getDef(), end->getDef()->getUse()[0],
//                         bb);
//     auto cmpDst = new Operand(new TemporarySymbolEntry(
//         TypeSystem::boolType, SymbolTable::getLabel()));
//     new CmpInstruction(CmpInstruction::GE, cmpDst, times_op, limit, bb);

//     new CondBrInstruction(succ, pred, cmpDst, cmpbb);
// }

// // Fix: 太大可能会溢出
// void LoopStrengthReduction::updateBlock_ari(int opcode, BasicBlock *bb,
//                                             int start, int step, Operand
//                                             *end) {
//     print("======== updateBlock_ari ========");
//     Operand *start_op =
//         new Operand(new ConstantSymbolEntry(TypeSystem::intType, start));
//     Operand *step_op =
//         new Operand(new ConstantSymbolEntry(TypeSystem::intType, step));
//     Operand *times_op = nullptr;
//     Operand *finalValue = nullptr;
//     if (start == 0 && step == 1) {
//         times_op = end;
//         finalValue = end;
//     } else {
//         switch (opcode) {
//             Operand *dst;
//         case CmpInstruction::L:
//             // times = (end - start) / step;

//             dst = new Operand(new TemporarySymbolEntry(
//                 TypeSystem::intType, SymbolTable::getLabel()));
//             new BinaryInstruction(BinaryInstruction::SUB, dst, end, start_op,
//                                   bb);

//             times_op = new Operand(new TemporarySymbolEntry(
//                 TypeSystem::intType, SymbolTable::getLabel()));
//             new BinaryInstruction(BinaryInstruction::SUB, times_op, dst,
//                                   step_op, bb);

//             finalValue = end;
//             break;
//         case CmpInstruction::G:
//             // times = (start - end) / step;
//             dst = new Operand(new TemporarySymbolEntry(
//                 TypeSystem::intType, SymbolTable::getLabel()));
//             new BinaryInstruction(BinaryInstruction::SUB, dst, start_op, end,
//                                   bb);

//             times_op = new Operand(new TemporarySymbolEntry(
//                 TypeSystem::intType, SymbolTable::getLabel()));
//             new BinaryInstruction(BinaryInstruction::SUB, times_op, dst,
//                                   step_op, bb);

//             finalValue = end;
//             break;
//         default:
//             print("invelid opcode");
//             return;
//         }
//     }

//     auto cond =
//         dynamic_cast<CondBrInstruction *>(findinst(Instruction::COND, bb));
//     if (end->getDef()->isLoad()) {
//         new LoadInstruction(end->getDef()->getDef(),
//         end->getDef()->getUse()[0],
//                             bb);
//     }

//     auto condbb = bb->getPredBlock()[0];
//     auto cond_prev =
//         dynamic_cast<CondBrInstruction *>(findinst(Instruction::COND,
//         condbb));
//     print("cond done.");
//     Operand *terminal = nullptr;
//     auto two = new Operand(new ConstantSymbolEntry(TypeSystem::intType, 2));
//     auto one = new Operand(new ConstantSymbolEntry(TypeSystem::intType, 1));
//     auto zero = new Operand(new ConstantSymbolEntry(TypeSystem::intType, 0));

//     BasicBlock *trueBranch = bb;
//     BasicBlock *falseBranch = new BasicBlock(bb->getParent());
//     auto cmpbb = new BasicBlock(bb->getParent());
//     Operand *ld;
//     if (end->getDef()->isLoad()) {
//         ld = new Operand(new TemporarySymbolEntry(TypeSystem::boolType,
//                                                   SymbolTable::getLabel()));
//         new LoadInstruction(ld, end->getDef()->getUse()[0], cmpbb);
//     }
//     auto mod2dst = new Operand(
//         new TemporarySymbolEntry(TypeSystem::intType,
//         SymbolTable::getLabel()));
//     new BinaryInstruction(BinaryInstruction::MOD, mod2dst, ld, two, cmpbb);
//     auto cmpDst = new Operand(new TemporarySymbolEntry(
//         TypeSystem::boolType, SymbolTable::getLabel()));
//     new CmpInstruction(CmpInstruction::E, cmpDst, mod2dst, zero, cmpbb);
//     new CondBrInstruction(trueBranch, falseBranch, cmpDst, cmpbb);

//     for (auto b : bb->getPredBlock()) {
//         cmpbb->addPred(b);
//         b->addSucc(cmpbb);
//     }
//     bb->getPredBlock().clear();
//     cmpbb->addSucc(bb);
//     cmpbb->addSucc(falseBranch);
//     bb->addPred(cmpbb);
//     falseBranch->addPred(cmpbb);
//     cond_prev->setTrueBranch(cmpbb);

//     for (auto pair : AccumInstEleVec) {
//         /* 处理含取模的累加 */
//         auto num = new Operand(
//             new ConstantSymbolEntry(TypeSystem::intType, NUMSECOND));
//         auto dst_div = new Operand(new TemporarySymbolEntry(
//             TypeSystem::intType, SymbolTable::getLabel()));
//         new BinaryInstruction(BinaryInstruction::DIV, dst_div, times_op, num,
//                               bb);
//         auto dst_left = new Operand(new TemporarySymbolEntry(
//             TypeSystem::intType, SymbolTable::getLabel()));
//         new BinaryInstruction(BinaryInstruction::DIV, dst_left, num, two,
//         bb);

//         Operand *dst_srem;
//         Operand *acc = zero;
//         for (int i = 0; i < NUMSECOND; i++) {

//             // 累加变量
//             auto addr = new Operand(new TemporarySymbolEntry(
//                 TypeSystem::intType, SymbolTable::getLabel()));
//             new BinaryInstruction(BinaryInstruction::DIV, addr, times_op,
//             two,
//                                   bb);

//             auto addr2 = new Operand(new TemporarySymbolEntry(
//                 TypeSystem::intType, SymbolTable::getLabel()));
//             new BinaryInstruction(BinaryInstruction::SUB, addr2, times_op,
//             one,
//                                   bb);

//             auto addr3 = new Operand(new TemporarySymbolEntry(
//                 TypeSystem::intType, SymbolTable::getLabel()));
//             new BinaryInstruction(BinaryInstruction::MUL, addr3, addr, addr2,
//                                   bb);

//             auto def = (pair.second).first->getDef();
//             print((pair.second).first->toStr());
//             Operand *opd;
//             if (def->isPhi()) {
//                 for (auto src :
//                      dynamic_cast<PhiInstruction *>(def)->getSrcs()) {
//                     if (src.first != bb) {
//                         opd = src.second;
//                     }
//                 }
//             } else if (def->isLoad()) {
//                 opd = (pair.second).first;
//             }
//             // pair.first->getDef()->getUse_v().clear();
//             new BinaryInstruction(pair.first->getOpcode(),
//             pair.first->getDef(),
//                                   opd, addr3, bb);
//             terminal = pair.first->getDef();
//         }
//     }
//     // if (end->getDef()->isLoad()) {
//     //     new LoadInstruction(end->getDef()->getDef(),
//     //     end->getDef()->getUse()[0],
//     //                         falseBranch);
//     // }
//     auto ld2 = new Operand(new TemporarySymbolEntry(TypeSystem::boolType,
//                                                     SymbolTable::getLabel()));
//     new LoadInstruction(ld2, end->getDef()->getUse()[0], falseBranch);
//     Operand *terminal2 = new Operand(
//         new TemporarySymbolEntry(TypeSystem::intType,
//         SymbolTable::getLabel()));
//     for (auto pair : AccumInstEleVec) {
//         /* 处理含取模的累加 */

//         // 累加变量

//         auto addr2 = new Operand(new TemporarySymbolEntry(
//             TypeSystem::intType, SymbolTable::getLabel()));
//         new BinaryInstruction(BinaryInstruction::SUB, addr2, ld2, one,
//                               falseBranch);

//         auto addr = new Operand(new TemporarySymbolEntry(
//             TypeSystem::intType, SymbolTable::getLabel()));
//         new BinaryInstruction(BinaryInstruction::DIV, addr, addr2, two,
//                               falseBranch);

//         auto addr3 = new Operand(new TemporarySymbolEntry(
//             TypeSystem::intType, SymbolTable::getLabel()));
//         new BinaryInstruction(BinaryInstruction::MUL, addr3, addr, ld2,
//                               falseBranch);

//         auto def = (pair.second).first->getDef();
//         print((pair.second).first->toStr());
//         Operand *opd;
//         if (def->isPhi()) {
//             for (auto src : dynamic_cast<PhiInstruction *>(def)->getSrcs()) {
//                 if (src.first != bb) {
//                     opd = src.second;
//                 }
//             }
//         } else if (def->isLoad()) {
//             opd = (pair.second).first;
//         }
//         // pair.first->getDef()->getUse_v().clear();
//         new BinaryInstruction(pair.first->getOpcode(), terminal2, opd, addr3,
//                               falseBranch);
//         // terminal = pair.first->getDef();
//     }
//     print("AccumInstEleVec done.");

//     // new BinaryInstruction(BinaryInstruction::MOD, bin_mod->getDef(),
//     // terminal,
//     //                       bin_mod->getUse()[1], bb);
//     // auto sum = new Operand(
//     //     new TemporarySymbolEntry(TypeSystem::intType,
//     //     SymbolTable::getLabel()));
//     new BinaryInstruction(BinaryInstruction::MOD,
//                           bin_mod->getDef()->getUse()[1]->getDef(),
//                           terminal2, bin_mod->getUse()[1], falseBranch);

//     // // 控制变量
//     // BinaryInstruction *loopControllerFinalVal = new BinaryInstruction(
//     //     BinaryInstruction::ADD, loopController, finalValue, zero, bb);
//     // cond -> uncond
//     BasicBlock *endbb =
//         dynamic_cast<CondBrInstruction *>(cond)->getFalseBranch();
//     // auto tmp2 = new BasicBlock(bb->getParent());
//     new UncondBrInstruction(endbb, bb);

//     new UncondBrInstruction(endbb, falseBranch);

//     // auto tmp = new BasicBlock(bb->getParent());
//     // new UncondBrInstruction(tmp, falseBranch);
//     // new UncondBrInstruction(endbb, tmp);
//     // falseBranch->addSucc(tmp);
//     // tmp->addPred(falseBranch);
//     // tmp->addSucc(endbb);
//     // endbb->addPred(tmp);

//     // new UncondBrInstruction(endbb, tmp2);
//     // bb->getSuccBlock().clear();
//     // bb->addSucc(tmp2);
//     // tmp2->addPred(bb);
//     // tmp2->addSucc(endbb);
//     // endbb->addPred(tmp2);

//     // std::cout << "bin_mod->getDef() = " << bin_mod->getDef()->toStr() << ""
//     //           << terminal->getUse().size() << "\n";
//     print("bin_mod->getDef() = " << bin_mod->getDef()->toStr() << " "
//                                  << terminal->getUse().size());
//     // if (bin_mod->getDef()->getUse()[1]->isPhi()) {
//     //     auto phi_t =
//     //         dynamic_cast<PhiInstruction*>(bin_mod->getDef()->getUse()[1]);
//     //     // std::cout << "8888 " << phi_t->getDef()->toStr() << "\n";
//     //     // phi_t->addSrc(tmp, sum);
//     //     // phi_t->addSrc(falseBranch, sum);
//     //     // phi_t->addSrc(tmp2, terminal);
//     // }

//     for (auto inst : OldInstrs) {
//         bb->remove(inst);
//         delete inst;
//     }
//     OldInstrs.clear();

//     // 维护前驱后继
//     auto succs = bb->getSuccBlock();
//     if (std::find(succs.begin(), succs.end(), bb) != succs.end())
//         bb->removeSucc(bb);

//     auto preds = bb->getPredBlock();
//     if (std::find(preds.begin(), preds.end(), bb) != preds.end())
//         bb->removePred(bb);
//     done = 1;
//     std::cout << "yes lsr-ari done.\n";
// }
