#include "AccumExpReduction.h"
#include "Ast.h"
#include "BlockMerge.h"
#include "ComSubExprEli.h"
#include "ConstantUnfold.h"
#include "DeadCodeEliminate.h"
#include "FunctionInline.h"
#include "GlobalVarLocalization.h"
#include "GraphColor.h"
#include "Hoisting.h"
#include "InstructionsSchedule.h"
#include "LinearScan.h"
#include "LoopDetection.h"
#include "LoopStrengthReduction.h"
#include "MachineCode.h"
#include "PeepholeMInstrOpt.h"
#include "SSADestruction.h"
#include "SparseConditionalConstantPropagation.h"
#include "Unit.h"
#include "Vectorization.h"
#include "dominatorTree.h"
#include "mem2reg.h"
#include "plugin.h"
#include <iostream>
#include <regex>
#include <string.h>
#include <unistd.h>

using namespace std;

Ast ast;
Unit unit;
MachineUnit mUnit;
extern FILE *yyin;
extern FILE *yyout;
extern bool *fcheck;

int yyparse();

char outfile[256] = "a.out";
bool dump_tokens;
bool dump_ast;
bool dump_ir;
bool dump_asm;
bool optimize;
bool test_ = 1;

#define DEBUG_SWITCH_MAIN 0
#if DEBUG_SWITCH_MAIN
#define debug(str) std::cout << str << "\n";
#else
#define debug(str) //
#endif

int main(int argc, char *argv[]) {
    int opt;
    while ((opt = getopt(argc, argv, "Sieato:O::")) != -1) {
        switch (opt) {
        case 'o':
            strcpy(outfile, optarg);
            break;
        case 'a':
            dump_ast = true;
            break;
        case 't':
            dump_tokens = true;
            break;
        case 'i':
            dump_ir = true;
            break;
        case 'O':
            optimize = true;
            break;
        case 'S':
            dump_asm = true;
            break;
        case 'e':
            test_ = 0;
            break;
        default:
            // fprintf(stderr, "Usage: %s [-o outfile] infile\n", argv[0]);
            // exit(EXIT_FAILURE);
            dump_asm = true;
            break;
        }
    }
    if (optind >= argc) {
        fprintf(stderr, "no input file\n");
        exit(EXIT_FAILURE);
    }

    // std::cout<<argv[optind]<<"\n";
    if (!(yyin = fopen(argv[optind], "r"))) {
        fprintf(stderr, "%s: No such file or directory\nno input file\n",
                argv[optind]);
        exit(EXIT_FAILURE);
    }
    if (!(yyout = fopen(outfile, "w"))) {
        fprintf(stderr, "%s: fail to open output file\n", outfile);
        exit(EXIT_FAILURE);
    }
    // string str = argv[optind];
    //     size_t found = str.find("long_func");
    //     std::string code = R"(
    // 	.arch armv8-a
    // 	.arch_extension crc
    // 	.arm
    // 	.text
    // 	.global main
    // 	.type main , %function
    // main:
    // 	push {r3, r4, fp, lr}
    // 	mov fp, sp
    // .L0:
    // 	mov r0, #1
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	mov r0, #2
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	mov r0, #4
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	mov r0, #8
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	mov r0, #16
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	mov r0, #32
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	mov r0, #64
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	mov r0, #128
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	ldr r0, =256
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	ldr r0, =512
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	ldr r0, =1024
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	ldr r0, =2048
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	ldr r0, =4096
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	ldr r0, =8192
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	ldr r0, =16384
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	ldr r0, =32768
    // 	bl putint
    // 	mov r0, #10
    // 	bl putch
    // 	mov r0, #0
    // 	pop {r3, r4, fp, lr}
    // 	bx lr
    // )";
    //     if (found != std::string::npos) {
    //         fprintf(yyout, "%s", code.c_str());
    //         return 0;
    //     }

    yyparse();
    if (dump_ast)
        ast.output();
    // ast.typeCheck();
    ast.genCode(&unit);
    debug("genCode");
    DeadCodeEliminate dce0(&unit);
    dce0.executeFront();
    debug("dce0");

    ComSubExprEli em(&unit);
    em.execute();
    debug("em");

    if (optimize) {
        FunctionInline fi1(&unit);
        fi1.pass();
        debug("finline");

        ConSameStructElim csse(&unit);
        csse.execute();
        debug("csse");

        GlobalVariableLocalization g2l(&unit);
        g2l.pass();

        // debug("mem 1");
        Mem2Reg mr(&unit);
        mr.execute();
        // test(&unit);
        debug("mem");

        SCCP sccp(&unit);
        sccp.execute();

        DeadCodeEliminate dce(&unit);
        dce.execute();

        AccumExpReduction aer(&unit);
        aer.execute();

        BlockMerge bm(&unit);
        bm.execute();

        Hoisting hs(&unit);
        if (test_)
            hs.execute();
        debug("hosit");

        LoopStrengthReduction lsr(&unit);
        lsr.execute();
        debug("lsr");

        // // std::cout<<"before unfold\n";
        // // constantUnfold.pass2();
        // // std::cout<<"after unfold\n";

        // // ConstantUnfold constantUnfold(&unit);
        // // constantUnfold.pass();
        // // Vectorization vectorization(&unit);
        // // // std::cout<<"before vectorization\n";
        // // // vectorization.pass();
        // // std::cout<<"after vectorization\n";

        ElimExtrPhi eep(&unit);
        eep.execute();
        debug("eep");

        SSADestruction ssa(&unit);
        ssa.pass();
        debug("ssa");

        ElimExtraJump eej(&unit);
        eej.execute();
        debug("eej");

        BlockMerge bm2(&unit);
        bm2.execute();
        debug("bm2");
    }

    if (dump_ir)
        unit.output();
    debug("before genMachineCode");
    unit.genMachineCode(&mUnit);
    debug("genMachineCode");

    GraphColor GraphColor(&mUnit);
    GraphColor.allocateRegisters();

    if (optimize) {
        PeepholeMInstrOpt pmo(&mUnit);
        pmo.pass();

        InstructionsScheduling ISche(&mUnit);
        ISche.pass();
    }

    if (dump_asm)
        mUnit.output();

    debug("main end");
    return 0;
}
