#include "plugin.h"

#define DEBUG_SWITCH_PLUGIN 0
#if DEBUG_SWITCH_PLUGIN
#define printp(str) \
    std::cout << str << "\n";
#else
#define printp(str) //
#endif

void ElimExtraJump::execute() {
    // 遍历所有的Function
    printp("==========start ElimExtraJump=========");
    std::vector<Function*> func = unit->get_functions();
    for(auto f = func.begin(); f != func.end(); f++){
        printp("=========start func ElimExtraJump=========");
        printp((*f)->getSymPtr()->toStr());
        execute(*f);
        printp("=========end func ElimExtraJump=========");
    }
    printp("==========end ElimExtraJump=========");
}

void ElimExtraJump::execute(Function*f) {
    // 遍历所有的块, 如果块内只有一条无条件跳转就删除对应的块和无条件跳转语句
    // 遍历所有的条件跳转和无条件跳转语句, 存起来, 每次遍历到对应的需要消除的块时就将这些集合中的块进行更新
    BlockToCondMap.clear();
    BlockToUncondMap.clear();
    deleteblocks.clear();
    ExtraToJump.clear();
    for (auto it = f->begin(); it != f->end(); it++) {
        BasicBlock *cur = *it;
        // 首先进行判断该block当中是否只有一条无条件跳转语句
        bool isExtr = isExtrBlock(cur);
        // 如果不是的话, 接下来遍历对应的block当中所有的条件语句
        if(!isExtr){
            collectAllCondAndUncond(cur);
        } 
    }
    // FIX: 可能后续进行块的遍历
    for(auto it: ExtraToJump){
        printp("==========start replace=========");
        printp("需要进行替换的块号: "<<it.first->getNo());
        replaceExtrBlock(it.first);
        printp("==========end replace=========");
        deleteblocks.push_back(it.first);
    }
    // FIX: 将删除块的操作放在前面
    ElimAllExtrBlock();
}

bool ElimExtraJump::isExtrBlock(BasicBlock* block){
    for (auto i = block->begin(); i != block->end(); i = i->getNext()) {
        if(i->isUncond() && i->getNext() == block->end()){
            UncondBrInstruction* uncond = dynamic_cast<UncondBrInstruction* >(i);
            BasicBlock* jumpBlock = uncond->getBranch();
            while(ExtraToJump.count(jumpBlock)){
                jumpBlock = ExtraToJump[jumpBlock];
            }
            ExtraToJump[block] = jumpBlock;
            printp("需要将块: "<<block->getNo()<<" ==> 替换为: "<<jumpBlock->getNo());
            for(auto it: ExtraToJump){
                if(it.second == block){
                    ExtraToJump[it.first] = jumpBlock;
                }
            }
            return true;
        }
        return false;
    }
    return false;
}

void ElimExtraJump::collectAllCondAndUncond(BasicBlock* block){
    for (auto i = block->begin(); i != block->end(); i = i->getNext()) {
        if(i->isCond()){
            CondBrInstruction* cond = dynamic_cast<CondBrInstruction* >(i);
            BasicBlock* true_branch = cond->getTrueBranch();
            BasicBlock* false_branch = cond->getFalseBranch();

            if(BlockToCondMap.count(true_branch)){
                std::vector<CondBrInstruction* >* list = BlockToCondMap[true_branch];
                list->push_back(cond);
            }else{
                std::vector<CondBrInstruction* >* list = new std::vector<CondBrInstruction* >();
                list->push_back(cond);
                BlockToCondMap[true_branch] = list;
            }

            if(BlockToCondMap.count(false_branch)){
                std::vector<CondBrInstruction* >* list = BlockToCondMap[false_branch];
                list->push_back(cond);
            }else{
                std::vector<CondBrInstruction* >* list = new std::vector<CondBrInstruction* >();
                list->push_back(cond);
                BlockToCondMap[false_branch] = list;
            }
        } else if(i->isUncond()){
            UncondBrInstruction* uncond = dynamic_cast<UncondBrInstruction* >(i);
            BasicBlock* branch = uncond->getBranch();
            if(BlockToUncondMap.count(branch)){
                std::vector<UncondBrInstruction* >* list = BlockToUncondMap[branch];
                list->push_back(uncond);
            }else{
                std::vector<UncondBrInstruction* >* list = new std::vector<UncondBrInstruction* >();
                list->push_back(uncond);
                BlockToUncondMap[branch] = list;
            }
        }
    }
}

void ElimExtraJump::replaceExtrBlock(BasicBlock* block){
    if(BlockToCondMap.count(block)){
        std::vector<CondBrInstruction* >* list = BlockToCondMap[block];
        for(auto cond: *list){
            BasicBlock* true_branch = cond->getTrueBranch();
            BasicBlock* false_branch = cond->getFalseBranch();
            if(true_branch == block){
                cond->setTrueBranch(ExtraToJump[block]);
                printp("将cond中的块: "<<true_branch->getNo()<<" ==> 更新为: "<<ExtraToJump[block]->getNo());
            }
            if(false_branch == block){
                cond->setFalseBranch(ExtraToJump[block]);
                printp("将cond中的块: "<<false_branch->getNo()<<" ==> 更新为: "<<ExtraToJump[block]->getNo());
            }
        }
    }

    if(BlockToUncondMap.count(block)){
        std::vector<UncondBrInstruction* >* list = BlockToUncondMap[block];
        for(auto uncond: *list){
            BasicBlock* branch = uncond->getBranch();
            if(branch == block){
                uncond->setBranch(ExtraToJump[block]);
                printp("将uncond中的块: "<<branch->getNo()<<" ==> 更新为: "<<ExtraToJump[block]->getNo());
            }
        }
    }

    // printp("开始删块: "<<block->getNo());
    // block->deleteAllRelation();
    // block->getParent()->remove(block);
    // printp("结束删块: "<<block->getNo());
}

void ElimExtraJump::ElimAllExtrBlock() {
    printp("开始删块: ");
    for(auto block: deleteblocks) {
        printp("删除块号: "<<block->getNo());
        Function* f = block->getParent();
        //  A->B->C->D
        // for(auto prevbb: block->getPrevbb()){
        //     prevbb->removeSucc(block);
        //     block->removePred(block);
        // }
        // for(auto succbb: block->getSuccBlock()){
        //     succbb->removePred(block);
        //     block->removeSucc(succbb);
        // }
        block->maintainAllRelation();
        block->deleteAllRelation();
        if(block == f->getEntry()){
            f->setEntry(ExtraToJump[block]);
            printp("修改entry: "<<block->getNo()<<" => "<<ExtraToJump[block]->getNo());
        }
        f->remove(block);
    }
}

#define DEBUG_SWITCH_PLUGIN_CSSE 0
#if DEBUG_SWITCH_PLUGIN_CSSE
#define printcsse(str) \
    std::cout << str << "\n";
#else
#define printcsse(str) //
#endif

void ConSameStructElim::execute() {
    // 遍历所有的Function
    printcsse("==========start ConSameStructElim=========");
    std::vector<Function*> func = unit->get_functions();
    for(auto f = func.begin(); f != func.end(); f++){
        printcsse("=========start func ConSameStructElim=========");
        printcsse((*f)->getSymPtr()->toStr());
        execute(*f);
        printcsse("=========end func ConSameStructElim=========");
    }
    printcsse("==========end ConSameStructElim=========");
}

void ConSameStructElim::execute(Function*f) {
    // 遍历块当中所有参数的指令流, 得到首个参数的指令流, 判断所有参数的指令流是否相同(首先判断除法, 后续可以加其它指令类型)
    instrflow.clear();
    allParams.clear();
    firstOperand = nullptr;
    setFlag(true);
    setFirstOperandSet(false);
    FunctionType *functype = dynamic_cast<FunctionType*>(f->getSymPtr()->getType());
    int num = functype->getParamSum();
    if(num < ParamNum) {
        setFlag(false);
        return;
    }
    // 得到函数当中所有的参数:
    getAllParamsFunction(f);
    // 得到首个参数的指令流:
    std::vector<std::string> paraminstnameflow;
    if(getFirstOperandSet()) {
        // 将首个参数的指令流进行集中
        paraminstnameflow = getParamInstructionFlow(f, firstOperand);
    }
    if(!paraminstnameflow.empty()){
        for(auto name: paraminstnameflow){
            printcsse("first param instr name: "<< name);
        }
    }
    // 对每一个参数进行判断指令流是否相同:
    if(allParams.size() < ParamNum){
        setFlag(false);
        return;
    }
    for(auto it: allParams){
        instrflow.clear();
        if(judgeSomeInstructionFlow(f, it, paraminstnameflow) == false){
            setFlag(false);
            break;
        }
    }
    // 如果指令流相同那么就进行参数运算合并:
    if(getFlag()){
        f->setcsseflag();
        mergeParamsInstruction(f);
    }
}

void ConSameStructElim::getAllParamsFunction(Function* f) {
    printcsse("==========start getAllParamsFunction=========");
    for (auto it = f->begin(); it != f->end(); it++) {
        for (auto i = (*it)->begin(); i != (*it)->end(); i = i->getNext()) {
            auto Uselist = i->getUse();
            for(auto op = Uselist.begin(); op != Uselist.end(); op++){
                if((*op)->isParam() && !allParams.count(*op)){
                    allParams.insert(*op);
                    if(!getFirstOperandSet()){
                        firstOperand = *op;
                        setFirstOperandSet(true);
                        printcsse("Set First Operand: "<<(*op)->getSymbolEntry()->toStr());
                    }
                    printcsse("Add Param Operand: "<<(*op)->getSymbolEntry()->toStr());
                }
            }
        }
    }
    printcsse("==========end getAllParamsFunction=========");
}

std::vector<std::string> ConSameStructElim::getParamInstructionFlow(Function*f, Operand *op) {
    printcsse("==========start getParamInstructionFlow=========");
    // 其中一种方式是遍历所有的指令, 看指令当中包含param operand的情况, 开销比较大, 但是指令顺序比较准确;
    // 另一种方式是遍历param operand->getuse -> 得到alloca operand->getUse getUse的每一条指令->getDef ->getUse, 
    // 关键是确定指令的顺序, 倾向于确定指令之后进行排序, 排序的方式是看在block当中指令的位置
    std::vector<Operand*> worklist;
    worklist.push_back(op);
    while(!worklist.empty()){
        Operand* curOp = worklist.back();
        worklist.pop_back();
        auto Uselist = curOp->getUse();
        for(auto it = Uselist.begin(); it != Uselist.end(); it++){
            // 将指令添加进入instrflow当中
            if(std::find(instrflow.begin(), instrflow.end(), (*it)) == instrflow.end()){
                instrflow.push_back(*it);
                printcsse("Add instrution: "<<(*it)->getInstName());
                // 将特定指令用到的寄存器加入到工作集当中
                if((*it)->isStore()) {
                    auto Op = dynamic_cast<StoreInstruction*>(*it)->getAddr();
                    worklist.push_back(Op);
                } else if((*it)->isLoad()) {
                    auto Op = dynamic_cast<LoadInstruction*>(*it)->getDef();
                    worklist.push_back(Op);
                } else if((*it)->isBin()) {
                    if(dynamic_cast<BinaryInstruction*>(*it)->getOpcode() == BinaryInstruction::DIV) {
                        auto Op = dynamic_cast<BinaryInstruction*>(*it)->getDef();
                        worklist.push_back(Op);
                    }
                }
            }
        }
    }
    // 将指令流进行排序得到顺序指令流
    std::vector<std::string> instrnameflow;
    instrnameflow = sortInstructionFlow(f);
    printcsse("==========end getParamInstructionFlow=========");
    return instrnameflow;
}

std::vector<std::string> ConSameStructElim::sortInstructionFlow(Function* f) {
    printcsse("==========start sortInstructionFlow=========");
    // 找到指令流当中的指令, 然后进行排序
    std::unordered_map<Instruction*, int> instrMap;
    std::vector<std::string> instnamevec;
    for (auto inst: instrflow){
        instrMap[inst] = -1;
    }
    int count = 0;
    auto entry = f->getEntry();
    
    for (auto it = f->begin(); it != f->end(); it++) {
        for (auto i = (*it)->begin(); i != (*it)->end(); i = i->getNext()) {
            count++;
            if(instrMap.count(i)){
                std::string instname = i->getInstName();
                if(i->isBin()){
                    instname += std::to_string(dynamic_cast<BinaryInstruction*>(i)->getOpcode());
                }
                instnamevec.push_back(instname);
                printcsse("Get instr: "<< instname);
            }
        }
    }

    printcsse("==========end sortInstructionFlow=========");
    return instnamevec;
}

bool ConSameStructElim::judgeSomeInstructionFlow(Function*f, Operand* param, std::vector<std::string>& firstparamvec) {
    printcsse("==========start judgeSomeInstructionFlow=========");
    std::vector<std::string> paramInstVec;
    paramInstVec = getParamInstructionFlow(f, param);
    if(!paramInstVec.empty()){
        for(auto name: paramInstVec){
            printcsse("other param instr name: "<< name);
        }
    }
    for(auto i = 0; i < paramInstVec.size(); i++){
        if(i >= firstparamvec.size()){
            return false;
        }
        if(paramInstVec[i] != firstparamvec[i]){
            return false;
        }
    }
    printcsse("Param: "<<param->getSymbolEntry()->toStr()<<" has same");
    printcsse("==========end judgeSomeInstructionFlow=========");
    return true;
}

void ConSameStructElim::mergeParamsInstruction(Function* f) {
    printcsse("==========start mergeParamsInstruction=========");
    // 重排指令顺序, 将Add指令放在之前, 将DIV指令放在之后
    // 以下是正常指令的生成 alloca-load-store
    // 注意维护指令的use-def链
    auto entry = f->getEntry();
    std::vector<Instruction*> instremove;
    bool flag = false;
    for (auto i = entry->begin(); i != entry->end(); i = i->getNext()) {
        if(!flag && i->isLoad()){
            bool firstflag = false;
            Operand* beforeDst;
            for(auto op: allParams) {
                auto allocaop = dynamic_cast<StoreInstruction*>(op->getUse()[0])->getAddr();
                auto new_dst_se3 = new TemporarySymbolEntry(TypeSystem::intType,
                                                    SymbolTable::getLabel());
                auto load_dst = new Operand(new_dst_se3);
                auto newload = new LoadInstruction(load_dst, allocaop);
                allocaop->addUse(newload);
                entry->insertBefore(newload, i);
                if(firstflag) {
                    auto add_se = new TemporarySymbolEntry(TypeSystem::intType,
                                                    SymbolTable::getLabel());
                    auto add_dst = new Operand(add_se);
                    auto new_add = new BinaryInstruction(BinaryInstruction::ADD, add_dst, beforeDst, load_dst);
                    entry->insertBefore(new_add, i);
                    beforeDst = add_dst;
                    continue;
                }
                firstflag = true;
                beforeDst = load_dst;
            }
            auto div_se = new TemporarySymbolEntry(TypeSystem::intType,
                                            SymbolTable::getLabel());
            auto div_dst = new Operand(div_se);
            auto two = new Operand(new ConstantSymbolEntry(TypeSystem::intType, 2));
            auto new_div = new BinaryInstruction(BinaryInstruction::DIV, div_dst, beforeDst, two);
            entry->insertBefore(new_div, i);
            auto new_ret = new RetInstruction(div_dst);
            entry->insertBefore(new_ret, i);
            flag = true;
        } 
        if(flag){
            instremove.push_back(i);
        }
    }
    for(auto inst: instremove) {
        for(auto op: inst->getUse()) {
            op->removeUse(inst);
        }
        entry->remove(inst);
    }
    // 以下是参数指令的生成 直接add div, 实际不可行
    // auto entry = f->getEntry();
    // auto newbb = new BasicBlock(f);
    // f->setEntry(newbb);
    // bool flag = false;
    // Operand* beforeop;
    // std::cout<<"start add: \n";
    // for(auto op: allParams) {
    //     if(flag) {
    //         auto add_se = new TemporarySymbolEntry(TypeSystem::intType,
    //                                                 SymbolTable::getLabel());
    //         auto add_dst = new Operand(add_se);
    //         new BinaryInstruction(BinaryInstruction::ADD, add_dst, op, beforeop, newbb);
    //         beforeop = add_dst;
    //         continue;
    //     }   
    //     flag = true;
    //     beforeop = op;
    // }
    // std::cout<<"end add: \n";
    // auto div_se = new TemporarySymbolEntry(TypeSystem::intType,
    //                                         SymbolTable::getLabel());
    // auto div_dst = new Operand(div_se);
    // auto two = new Operand(new ConstantSymbolEntry(TypeSystem::intType, 2));
    // auto new_div = new BinaryInstruction(BinaryInstruction::DIV, div_dst, beforeop, two, newbb);
    // auto new_ret = new RetInstruction(div_dst, newbb);
    // std::cout<<"end ret: \n";
    // entry->deleteAllRelation();
    // f->remove(entry);
    printcsse("==========end mergeParamsInstruction=========");
}


#define DEBUG_SWITCH_PLUGIN_EEP 0
#if DEBUG_SWITCH_PLUGIN_EEP
#define printeep(str) \
    std::cout << str << "\n";
#else
#define printeep(str) //
#endif

void ElimExtrPhi::execute() {
    // 遍历所有的Function
    printeep("==========start ElimExtrPhi=========");
    std::vector<Function*> func = unit->get_functions();
    for(auto f = func.begin(); f != func.end(); f++){
        printeep("=========start func ElimExtrPhi=========");
        printeep((*f)->getSymPtr()->toStr());
        execute(*f);
        printeep("=========end func ElimExtrPhi=========");
    }
    printeep("==========end ElimExtrPhi=========");
}

void ElimExtrPhi::execute(Function* f) {
    // 先写简单的: 消除无用块带来的phi的分支
    // 然后处理比较复杂的情况: 对于相同分支的phi可以替换为一个phi
    phiBB = nullptr;
    PhiInstList.clear();
    ExtrPhiToOnlyPhi.clear();
    PhiStructList.clear();
    for (auto it = f->begin(); it != f->end(); it++) {
        // 找到当前块的所有phi指令
        PrePhiList.clear();
        phiBB = *it;
        for (auto i = (*it)->begin(); i != (*it)->end(); i = i->getNext()) {
            if(!i->isPhi()){
                // break;
                continue;
            }
            auto phi = dynamic_cast<PhiInstruction*>(i);
            printeep(phi->getDef()->toStr()<<" ");
            PrePhiList.push_back(phi);
        }
        // 判断当前phi指令是否存在无用的分支跳转并消除
        printeep("start process phi, block no: "<<(*it)->getNo());
        processPhi(f);
        // 找到相同分支的phi然后进行替换 phi1 <=> phi2  phi2->getUse()->替换为phi1
    }
    // 下面进行遍历所有的phi指令:
    bool flag = true;
    // 重复替换直到没有phi指令可以被替换
    while(flag){
        flag = false;
        for(auto it: PhiInstList){
            auto bb = it.first;
            std::vector<PhiInstruction* >* curphilist = PhiInstList[bb];
            // 判断每一个block是否有phi指令可以进行替换, 如果有则进行替换
            // 1. 存储phi指令的结构 2. 比较phi指令的结构 3. 替换可以被替换的phi指令
            if(hasPhiReplaced(*curphilist)){
                flag = true;
                // 需要更新一下当前的PhiInstList[bb]
                updatePhiInstList(bb);
            }
        }
    }
    // 最后一步进行处理只有一个分支的phi指令:
    elimNoUsePhi(f);
}

void ElimExtrPhi::elimNoUsePhi(Function* f) {
    printeep("==========start elimNoUsePhi=========");
    for(auto it: PhiInstList){
        auto bb = it.first;
        printeep("bb: "<<bb->getNo());
        std::vector<PhiInstruction* >* curphilist = PhiInstList[bb];
        std::vector<BasicBlock*> srcblocklist;
        std::vector<Operand*> oplist;
        std::unordered_set<int> bbbbbb;
        for(auto phi: *curphilist) {

            printeep("  phi: "<<phi->getDef()->toStr());
            srcblocklist.clear();
            oplist.clear();
            auto srclist = phi->getSrcs();
            for(auto src: srclist) {
                auto bb = src.first;
                auto op = src.second;
                auto blocklist = f->getBlockList();
                if(std::find(blocklist.begin(), blocklist.end(), bb) == blocklist.end()){
                    if(bbbbbb.count(bb->getNo()) == 0){
                        bbbbbb.insert(bb->getNo());
                        printeep("no find block: "<<bb->getNo());
                    }
                    continue;
                } else{
                    srcblocklist.push_back(bb);
                    oplist.push_back(op);
                }
            }
            if(srcblocklist.size() == 0) {
                phi->getParent()->remove(phi);
                printeep("remove phi had no useful block: "<<phi->getDef()->toStr());
            } 
            if(srcblocklist.size() == 1) {
                auto zero = new Operand(new ConstantSymbolEntry(TypeSystem::intType, 0));
                auto add_inst = new BinaryInstruction(BinaryInstruction::ADD, phi->getDef(), oplist[0], zero);
                phi->getParent()->insertBefore(add_inst, phi);
                phi->getParent()->remove(phi);
                printeep("remove phi had one useful block: "<<phi->getDef()->toStr());
            } 
        }
    }
    printeep("==========start elimNoUsePhi=========");
}

void ElimExtrPhi::updatePhiInstList(BasicBlock* bb) {
    printeep("==========start updatePhiInstList=========");
    // 找到当前块的所有phi指令
    PrePhiList.clear();
    for (auto i = bb->begin(); i != bb->end(); i = i->getNext()) {
        if(!i->isPhi()){
            continue;
            // break;
        }
        auto phi = dynamic_cast<PhiInstruction*>(i);
        PrePhiList.push_back(phi);
    }
    // 判断当前phi指令是否存在无用的分支跳转并消除
    printeep("start process phi, block no: "<<bb->getNo());
    phiBB = bb;
    processPhi(bb->getParent());
    printeep("==========start updatePhiInstList=========");
}

bool ElimExtrPhi::hasPhiReplaced(std::vector<PhiInstruction*> phivec) {
    printeep("==========start hasPhiReplaced=========");
    bool flag = false;
    PhiStructList.clear();
    std::vector<PhiInstruction*> removelist;
    for(auto phi: phivec) {
        PhiStruct* curStruct = new PhiStruct(phi);
        if(PhiStructList.empty()) {
            PhiStructList.push_back(curStruct);
            continue;
        } 
        bool getSame = false;
        for(auto i = 0; i < PhiStructList.size(); i++) {
            if(*curStruct == *PhiStructList[i]) {
                // 如果检测到了相同的phi指令结构就可以进行替换, 这里先要把要替换的phi指令准备好, 之后再进行替换
                flag = true;
                if(ExtrPhiToOnlyPhi.count(PhiStructList[i]->phiDef)) {
                    ExtrPhiToOnlyPhi[phi->getDef()] = ExtrPhiToOnlyPhi[PhiStructList[i]->phiDef];
                } else {
                    ExtrPhiToOnlyPhi[phi->getDef()] = PhiStructList[i]->phiDef;
                }
                removelist.push_back(phi);
                printeep("Get same Struct Phi: "<< phi->getDef()->toStr() << " <=> "<< PhiStructList[i]->phiDef->toStr());
                getSame = true;
                break;
            }
        }
        if(!getSame) {
            PhiStructList.push_back(curStruct);
        }
    }
    for(auto phi: removelist) {
        // 这里将当前可以替换的phi指令的def全部进行替换
        phi->getDef()->replaceAllUseWith(ExtrPhiToOnlyPhi[phi->getDef()]);
        phi->getDef()->removeDef(phi);
        phi->getParent()->remove(phi);
    }
    printeep("==========end hasPhiReplaced=========");
    return flag;
}

void ElimExtrPhi::processPhi(Function* f) {
    printeep("==========start processPhi=========");
    std::unordered_set<int> bbbbbb;
    std::vector<BasicBlock*> srcblocklist;
    std::vector<Operand*> oplist;
    std::vector<PhiInstruction* >* newPhiList = new std::vector<PhiInstruction* >();
    for(auto phi: PrePhiList) {
        srcblocklist.clear();
        oplist.clear();
        auto srclist = phi->getSrcs();
        for(auto src: srclist) {
            auto bb = src.first;
            auto op = src.second;
            auto blocklist = f->getBlockList();
            if(std::find(blocklist.begin(), blocklist.end(), bb) == blocklist.end()){
                if(bbbbbb.count(bb->getNo()) == 0){
                    bbbbbb.insert(bb->getNo());
                    printeep("no find block: "<<bb->getNo());
                }
                continue;
            } else{
                srcblocklist.push_back(bb);
                oplist.push_back(op);
                // printeep("Add src block: "<<bb->getNo());
                // printeep("Add src op: "<<op->getSymbolEntry()->toStr());
            }
        }
        if(srcblocklist.size() == 0) {
            phi->getParent()->remove(phi);
            // printeep("remove phi had no useful block: "<<phi->getDef()->toStr());
        } 
        else {
            phi->clearSrcs();
            for(auto i = 0; i < srcblocklist.size(); i++){
                phi->addSrc(srcblocklist[i], oplist[i]);
            }
            newPhiList->push_back(phi);
            // printeep("fix phi had more than one useful block: "<<phi->getDef()->toStr());
        }
    }
    printeep("phiBB: "<<phiBB->getNo());
    PhiInstList[phiBB] = newPhiList;
    printeep("==========end processPhi=========");
}